# from abc import abstractmethod
# from abc import ABC
# import inspect

from lmfit import Parameters
import numpy as np
from scipy.integrate import odeint
from ..spectrum import fi
import scipy
from copy import deepcopy
from scipy.linalg import lstsq
from ..general_model import GeneralModel
import numba as nb
# import ray
# import psutil

# num_cpus = psutil.cpu_count(logical=False)
# ray.init(num_cpus=num_cpus, ignore_reinit_error=True)

posv = scipy.linalg.get_lapack_funcs(('posv'))
# gels = scipy.linalg.get_lapack_funcs(('gels'))


# import matplotlib.pyplot as plt
# from sympy import Derivative, symbols, Function, Equality
# from console import Console


def get_xy(data, x0=None, x1=None):
    """Get x and y values for current exp data for range [x0, x1]."""
    start = 0
    end = data.shape[0]

    if x0 is not None and x1 is not None and x0 > x1:
        x0, x1 = x1, x0

    if x0 is not None:
        start = fi(data[:, 0], x0)

    if x1 is not None:
        end = fi(data[:, 0], x1) + 1

    # x = data[start:end, 0]
    # y = data[start:end, 1]

    # return x, y
    return data[start:end, :]


def get_xy_unpacked(data, x0=None, x1=None):
    """Get x and y values for current exp data for range [x0, x1]."""

    arr = get_xy(data, x0, x1)
    return arr[:, 0], arr[:, 1]


def OLS_ridge(A, B, alpha=0.000):
    """
    Fast least squares solver for X: AX=B by direct solve method with optional Tikhonov regularization.
    """

    ATA = A.T.dot(A)
    ATB = A.T.dot(B)

    if alpha != 0:
        ATA.flat[::ATA.shape[-1] + 1] += alpha

    # call the LAPACK function, direct solve method
    c, x, info = posv(ATA, ATB, lower=False,
                      overwrite_a=False,
                      overwrite_b=False)

    return x


@nb.vectorize(nopython=True)
def parallel_model(j, t, k):
    """Exponential model with simple implementation of heaviside function.
    j is initial population vector, t is time and k is rate constant"""
    if t >= 0:
        return j * np.exp(-t * k)
    else:
        return 0

# @vectorize(nopython=True)
# def fold_exp(t, k, fwhm):
#
#     w = fwhm / (2 * np.sqrt(np.log(2)))  # width
#
#     if w > 0:
#         return 0.5 * np.exp(k * (k * w * w / 4.0 - t)) * math_erfc(w * k / 2.0 - t / w)
#     else:
#         return np.exp(-t * k) if t >= 0 else 0


def target_1st_order(t, K, j, numerical=False):
    """ t - times vector,
        K - Transfer matrix, j - initial population vector"""

    if numerical:
        # return odeint(lambda c, _t: K.dot(c), j, t)
        raise NotImplementedError()

    # based on Ivo H.M. van Stokkum equation in doi:10.1016/j.bbabio.2004.04.011

    L, Q = np.linalg.eig(K)
    Q_inv = np.linalg.inv(Q)

    A2_T = Q * Q_inv.dot(j)  # Q @ np.diag(Q_inv.dot(j))

    C = np.heaviside(t[:, None], 1) * np.exp(t[:, None] * L[None, :])

    return C.dot(A2_T.T)

# abstract class, all models has to inherit from this class
class Model(object):

    name = '--model name--'

    def __init__(self, exps_data: list, ranges: [list, tuple] = None, varpro: bool = True,
                 exp_dep_params: list = None, n_spec: int = 2,
                 spec_names: [list, tuple] = None, spec_visible: dict = None,
                 weight_func=lambda res, y: res, **kwargs):
        """ TODO-->> exp_data is list of 2D ndaarrays """
        self.exps_data = list(exps_data) if isinstance(exps_data, (list, tuple)) else [exps_data]
        self.ranges = ranges  # list of lists or None for no ranges
        self.set_ranges(self.ranges)

        self.varpro = varpro  # use Variable Projection - efficient calculation of amplitudes
        self.exp_dep_params = exp_dep_params  # experiment dependent params
        self.n_spec = n_spec
        self.params = None
        self.weight_func = weight_func

        self.spec_names = spec_names
        if self.spec_names is None:
            self.spec_names = list('ABCDEFGHIJKLMNOPQRSTUVWXYZ')

        self.spec_visible = spec_visible  # dictionary
        self.exp_indep_params = None  # experimental independent parameters, will be set in init_params method
        self.equal_pars = {}  # hold a exp_dep_param as a key and list tuple of pairs of equal parameters as a value

        if self.spec_visible is None:
            self.spec_visible = [{name: True for name in self.spec_names[:self.n_spec]} for _ in range(len(self.exps_data))]

        self.param_names_dict = {}

        self.data_ids = None
        # if ray.is_initialized():
        #     self.data_ids = [ray.put(data) for data in self.exps_data]

    def set_ranges(self, ranges=None):
        if self.exps_data is None:
            return

        if ranges is None:
            self.ranges = [(None, None)] * len(self.exps_data)
            return

        # only one range, make it valid for all experiements
        if isinstance(ranges, (list, tuple)) and isinstance(ranges[0], (float, int)):
            self.ranges = [ranges] * len(self.exps_data)
        else:
            self.ranges = ranges

    def get_current_species_names(self):
        return self.spec_names[:self.n_spec]

    def update_model_options(self, **kwargs):
        for key, value in kwargs.items():
            if not hasattr(self, key):
                raise TypeError(f'Argument {key} is not valid.')
            self.__setattr__(key, value)

        if 'n_spec' in kwargs and 'spec_visible' not in kwargs:
            new_spec_visible = [{name: True for name in self.spec_names[:self.n_spec]} for _ in range(len(self.exps_data))]

            # update spec_visible and keep the old parameters
            for new_dict, old_dict in zip(new_spec_visible, self.spec_visible):
                for key, value in old_dict.items():
                    if key in new_dict:
                        new_dict[key] = value

            self.spec_visible = new_spec_visible

        if 'n_spec' in kwargs:
            if self.exp_dep_params is not None and self.exp_indep_params is not None:
                all_params = self.get_all_param_names().keys()
                exp_dep_pars_old = self.exp_dep_params

                self.exp_dep_params = self.default_exp_dep_params()
                for old_par in exp_dep_pars_old:
                    if old_par in all_params:
                        if old_par not in self.exp_dep_params:
                            self.exp_dep_params.append(old_par)
                for old_par in self.exp_indep_params:
                    if old_par in all_params:
                        if old_par in self.exp_dep_params:
                            self.exp_dep_params.remove(old_par)

            else:
                self.exp_dep_params = self.default_exp_dep_params()

        if len(kwargs) > 0:
            self.update_params()

    def update_params(self):
        """Update the parameters based on new model options."""

        old_params = deepcopy(self.params)

        self.init_params()

        if self.params is not None and old_params is not None:
            for key, par in old_params.items():
                if key in self.params:
                    new_par = self.params[key]
                    new_par.value = par.value
                    new_par.vary = par.vary
                    new_par.min = par.min
                    new_par.max = par.max
                    new_par.stderr = par.stderr
                    if hasattr(par, 'enabled'):  # enabled option is added for GUI text fields
                        new_par.enabled = par.enabled

        del old_params

    def init_params(self):
        """Initialized the parameters for current model options"""
        self.params = Parameters()

        if self.exps_data is None or not np.iterable(self.exps_data):
            raise ValueError('Experimental data that are not iterable')

    def simulate(self):
        """Simulates the data and returns the tuple of simulated traces and residuals as lists filled with ndarrays"""
        pass

    def residuals(self):
        """Efficient calculation of only residuals, can be optimized for varpro fitting."""
        pass

    def get_all_param_names(self):
        """returns dictionary of available parameters with keys as names and values as explanations"""
        return dict()

    # def get_available_params_names(self):
    #     return set()

    def is_rate_par(self, par):
        return par.startswith('k_')

    def is_amp_par(self, par):
        return par in self.spec_names

    def is_j_par(self, par):
        return par.startswith('_')

    def is_intercept(self, par):
        return par == 'intercept'

    def format_exp_par(self, par, i):
        return f'{par}_e{i}'

    def default_exp_dep_params(self):
        """Returns the set of default experiment-dependent parameters."""
        pars = self.spec_names[:self.n_spec]
        pars += ['intercept']
        return pars

    def get_ordered_values(self, type: str, exp_num=0):
        return [self.params[param].value for param in self.param_names_dict[exp_num][type]]

    def get_model_indep_params_list(self):
        return [self.params[name] for name in self.exp_indep_params]

    def get_model_dep_params_list(self):
        pars_list = []

        # find experiment-dependent parameters form 'all' field
        for i in range(len(self.exps_data)):
            exp_pars = []
            for exp_dep_par in self.exp_dep_params:
                for par in self.param_names_dict[i]['all']:
                    if par.startswith(exp_dep_par):
                        exp_pars.append(self.params[par])

            pars_list.append(exp_pars)

        return pars_list

    def fix_all_exp_dep_params(self, vary: bool, par_name: str):
        if par_name in self.exp_indep_params:
            return

        all_params = map(lambda d: d['all'], self.param_names_dict)

        exp_entry = list(filter(lambda lst: par_name in lst, all_params))[0]

        index = exp_entry.index(par_name)
        for exp_pars in map(lambda d: d['all'], self.param_names_dict):
            parameter = self.params[exp_pars[index]]
            parameter.vary = vary

    def set_all_spec_visible(self, visible: bool, species: str):
        if species not in self.spec_visible[0]:
            return

        for vis_exps in self.spec_visible:
            vis_exps[species] = visible

    def model_options(self):
        """Returns list of dictionaries of all additional options for a given model that
        will be added to fitwidget"""
        # return [dict(type=bool, name='option_name', value=True)]
        return [dict(type=bool, name='varpro', value=True,
                     description='Calculate amplitudes from data by OLS')]


class _InterceptVarProModel(Model):

    def __init__(self, *args, fit_intercept_varpro=True, **kwargs):
        super(_InterceptVarProModel, self).__init__(*args, **kwargs)

        self.fit_intercept_varpro = fit_intercept_varpro

        self.data_ids = [None] * len(self.exps_data)
        # views of original data based on used defined ranges
        self.data_views = [None] * len(self.exps_data)
        self.batch_size = 1

    def set_equal_param_each(self, par_name: str = '_Q_0', n: int = 2):
        """Fills the equal_pars dictionary for a given par_name so that each n subsequent
        experiments will share the same parameter. Eg. for 6 different experiment and
        parameter _Q_0 and n=2, the resulting pairs will be
        equal_pars['_Q_0'] = [(0, 1), (2, 3), (4, 5)] which means that there will be 3
        different _Q_0 parameters for these 3 experiment pairs. First two experiments
        will use _Q_0_e01 parameter, another two experiments parameter _Q_0_e23, etc."""

        n_exps = len(self.exps_data)
        if n_exps % n != 0:
            raise ValueError(f"Number of experiments is not divisible by {n}.")

        space = np.arange(0, n_exps).reshape((n_exps // n, n))
        pairs = [tuple(row) for row in space]
        self.equal_pars[par_name] = pairs
        self.update_params()

    def model_options(self):
        opts = super(_InterceptVarProModel, self).model_options()
        fit_intercept_varpro_opt = dict(type=bool, name='fit_intercept_varpro', value=True,
                                        description='Calculate intercept from data by OLS')

        return opts + [fit_intercept_varpro_opt]

    def _get_traces(self, t, ks, j, i):
        """i is the current exp index"""
        raise NotImplementedError()

    def add_params(self, param_set: list = None, dict_params: dict = None, j_dict: dict = None, rates_dict: dict = None,
                   amps_dict: dict = None, species_visible: [dict] = None, par_format=lambda name: name, **kwargs):
        pass

    def get_param_dicts(self):
        rates_dict = {f'k_{i + 1}': i for i in range(self.n_spec)}
        amps_dict = {name: i for i, name in enumerate(self.spec_names[:self.n_spec])}
        j_dict = {f'_{name}_0': i for i, name in enumerate(self.spec_names[:self.n_spec])}

        return dict(rates_dict=rates_dict, amps_dict=amps_dict, j_dict=j_dict)

    def init_params(self):
        """Initialize the parameters for current model options"""
        super(_InterceptVarProModel, self).init_params()

        av_params = self.get_all_param_names().keys()
        # self.exp_dep_params = self.default_exp_dep_params() if self.exp_dep_params is None else self.exp_dep_params
        self.exp_indep_params = []
        for par in av_params:  # keep the order
            if par not in self.exp_dep_params:
                self.exp_indep_params.append(par)

        kwargs = self.get_param_dicts()

        params_indep = dict(all=[], rates=[''] * len(kwargs['rates_dict']),
                            j=[''] * len(kwargs['j_dict']),
                            amps=[''] * len(kwargs['amps_dict']), intercept='')

        n = len(self.exps_data)

        # add experiment independent parameters
        self.add_params(self.exp_indep_params, params_indep, **kwargs)

        self.param_names_dict = [deepcopy(params_indep) for _ in range(n)]

        # setup equal parameter names
        equal_pars_dict = {}
        for par_name, pairs in self.equal_pars.items():
            equal_pars_dict[par_name] = ['' for _ in range(len(self.exps_data))]

            for pair in pairs:
                idx_name = ''.join([str(i) for i in pair])
                for idx in pair:
                    equal_pars_dict[par_name][idx] = idx_name

        # add experiment dependent parameters
        for i in range(n):
            def par_format(name):
                if name in equal_pars_dict:
                    idx = equal_pars_dict[name][i]
                    if idx != '':
                        return self.format_exp_par(name, idx)

                return self.format_exp_par(name, i)

            self.add_params(self.exp_dep_params, self.param_names_dict[i], **kwargs,
                            species_visible=self.spec_visible[i],
                            par_format=par_format)

    def get_rate_values(self, exp_num):
        return np.asarray([self.params[p].value for p in self.param_names_dict[exp_num]['rates']])

    # @staticmethod
    # @ray.remote
    # def _get_results_remote(*ts, rates=None, amps=None, visibles=None, js=None, batch=None, **kwds):
    #     pass

    # def put_data_to_shared_memory(self):
    #     for i, (data, x_range, par_names) in enumerate(zip(self.exps_data, self.ranges, self.param_names_dict)):
    #         d_view = get_xy(data, x0=x_range[0], x1=x_range[1])
    #         self.data_views[i] = d_view
    #         self.data_ids[i] = ray.put(d_view)  # put view to shared memory

    # def simulate(self, params=None):
    #     """Simulates the data and returns the list of simulated traces as ndarrays
    #         works for ray..."""
    #
    #     if params is not None:
    #         self.params = params
    #
    #     if self.exps_data is None or not np.iterable(self.exps_data):
    #         raise ValueError('No experimental data or data are not iterable')
    #
    #     if self.data_views[0] is None:
    #         self.put_data_to_shared_memory()
    #
    #     lstsq_intercept = self.fit_intercept_varpro and 'intercept' in self.exp_dep_params
    #
    #     exp_indep_amps = [par for par in self.exp_indep_params if self.is_amp_par(par)]
    #
    #     results_ids = []
    #
    #     n_exps = len(self.exps_data)
    #     space = np.arange(0, n_exps).reshape((n_exps // self.batch_size, self.batch_size))
    #     idxs = [tuple(row) for row in space]
    #
    #     for batch in idxs:
    #
    #         js = [np.asarray([self.params[p].value for p in self.param_names_dict[i]['j']]) for i in batch]
    #         rates = [self.get_rate_values(i) for i in batch]
    #         amps = [np.asarray([self.params[p].value for p in self.param_names_dict[i]['amps']]) for i in batch]
    #         data = [self.data_ids[i] for i in batch]
    #         visibles = [self.spec_visible[i] for i in batch]
    #         intercepts = [self.params[self.param_names_dict[i]['intercept']].value for i in batch]
    #
    #         results_ids.append(self._get_results_remote.remote(*data, rates=rates, amps=amps, intercepts=intercepts,
    #                                                            visibles=visibles,
    #                                                            js=js, batch=batch, sequential=True, varpro=self.varpro,
    #                                                            lstsq_intercept=lstsq_intercept, exp_indep_amps=exp_indep_amps))
    #
    #     fits = [None] * len(self.exps_data)
    #     residuals = [None] * len(self.exps_data)
    #     x_vals = [None] * len(self.exps_data)
    #
    #     while len(results_ids):
    #
    #         done_id, results_ids = ray.wait(results_ids)
    #         result = ray.get(done_id[0])
    #
    #         for i, fit, amps, intercept in result:
    #
    #             for j in range(len(amps)):
    #                 self.params[self.param_names_dict[i]['amps'][j]].value = amps[j]
    #
    #             self.params[self.param_names_dict[i]['intercept']].value = intercept
    #
    #             x_vals[i] = self.data_views[i][:, 0]
    #             y = self.data_views[i][:, 1]
    #
    #             res = self.weight_func(fit - y, y)  # residual, use original data
    #             fits[i] = fit
    #             residuals[i] = res
    #
    #     return x_vals, fits, residuals

    def simulate(self, params=None):
        """Simulates the data and returns the list of simulated traces as ndarrays"""

        if params is not None:
            self.params = params

        if self.exps_data is None or not np.iterable(self.exps_data):
            raise ValueError('No experimental data or data are not iterable')

        x_vals = []
        fits = []
        residuals = []

        lstsq_intercept = self.fit_intercept_varpro and 'intercept' in self.exp_dep_params

        exp_indep_amps = [par for par in self.exp_indep_params if self.is_amp_par(par)]

        for i, (data, x_range, par_names, visible) in enumerate(zip(self.exps_data, self.ranges, self.param_names_dict,
                                                                    self.spec_visible)):

            x, y = get_xy_unpacked(data, x0=x_range[0], x1=x_range[1])
            _y = y.copy()  # copy view of y, it may change, otherwise, original data would be changed
            x_vals.append(x)

            j = np.asarray([self.params[p].value for p in par_names['j']])
            rates = self.get_rate_values(i)

            traces = self._get_traces(x, rates, j, i)  # simulate

            if self.varpro:

                amps_params = [self.params[p] for p in par_names['amps']]

                # exp indep traces
                exp_dep_select = []
                exp_indep_select = []
                for key, visible in visible.items():
                    is_independent = key in exp_indep_amps
                    exp_indep_select.append(is_independent and visible)
                    exp_dep_select.append(not is_independent and visible)

                A = traces[:, exp_dep_select]  # select only visible species
                # add intercept as constant function

                fit = 0
                if lstsq_intercept:
                    A = np.hstack((A, np.ones_like(x)[:, None]))
                else:
                    fit = self.params[par_names['intercept']].value
                    _y -= fit

                if any(exp_indep_select):  # calculate traces for independent-exp amplitudes and add to fit
                    _amps = np.asarray([p.value for p, indep in zip(amps_params, exp_indep_select) if indep])
                    exp_indep_traces = traces[:, exp_indep_select].dot(_amps)  # add calculated traces
                    fit += exp_indep_traces
                    _y -= exp_indep_traces

                # solve the least squares problem, find the amplitudes of visible compartments based on data
                amps = OLS_ridge(A, _y, 0)  # A @ amps = y - A_fixed @ amps_fixes - intercept

                fit += A.dot(amps)  # calculate the fit and add it

                # update amplitudes and intercept
                if lstsq_intercept:
                    *amps, intercept = list(amps)
                    self.params[par_names['intercept']].value = intercept

                amp_names = [amp for amp, selected in zip(amps_params, exp_dep_select) if selected]
                for par, coef in zip(amp_names, amps):
                    par.value = coef

            else:
                amps = np.asarray([self.params[p].value for p in par_names['amps']])
                fit = traces.dot(amps)  # weight the simulated traces with amplitudes and calculate the fit

                if lstsq_intercept:
                    intercept = (y - fit).sum() / fit.shape[0]  # calculate intercept by least squares
                    fit += intercept
                    self.params[par_names['intercept']].value = intercept
                else:
                    fit += self.params[par_names['intercept']].value  # just add it to fit

            res = self.weight_func(fit - y, y)  # residual, use original data
            fits.append(fit)
            residuals.append(res)

        return x_vals, fits, residuals

    def residuals(self, params=None):
        _, _, residuals = self.simulate(params)

        # stack all residuals and return
        return np.hstack(residuals)


class SeqParModel(_InterceptVarProModel):

    name = 'Sequential/Parallel Model (1st order)'

    def __init__(self, *args, sequential=True, **kwargs):
        spec_names = list('ABCDEFGHIJKLMNOPQRSTUVWXYZ')  # use alphabet
        kwargs.update(spec_names=spec_names)
        super(SeqParModel, self).__init__(*args, **kwargs)

        self.sequential = sequential

    def model_options(self):
        opts = super(SeqParModel, self).model_options()
        sequential_opt = dict(type=bool, name='sequential', value=True, description='Use sequential model')

        return opts + [sequential_opt]

    def update_model_options(self, **kwargs):
        super(SeqParModel, self).update_model_options(**kwargs)

        # handle the individual model option changes / override the default implementation
        # also handles visible changes
        for exp_params, visible in zip(self.param_names_dict, self.spec_visible):
            pars_j = [self.params[name] for name in exp_params['j']]
            for i in range(len(pars_j)):
                pars_j[i].vary = False
                pars_j[i].value = 1 if i == 0 or not self.sequential else 0

            for amp, vis in zip((self.params[name] for name in exp_params['amps']), visible.values()):
                cond = not self.varpro and vis and amp.name not in self.exp_indep_params
                amp.vary = cond
                amp.enabled = cond
                if not vis:
                    amp.value = 0

            self.params[exp_params['intercept']].vary = not self.fit_intercept_varpro
            self.params[exp_params['intercept']].enabled = not self.fit_intercept_varpro

        if 'intercept' in self.exp_indep_params:
            self.params[self.param_names_dict[0]['intercept']].enabled = True

        for par_name in self.exp_indep_params:
            if self.is_amp_par(par_name):
                p = self.params[par_name]
                p.vary = True
                p.enabled = True

    def get_all_param_names(self):
        pars = {}
        pars.update({f'_{name}_0': f'Initial concentration of {name}' for name in self.spec_names[:self.n_spec]})
        pars.update({name: f'Amplitude of {name}' for name in self.spec_names[:self.n_spec]})  # amplitudes
        pars.update({f'k_{i+1}': f'Rate constant k_{i+1}' for i in range(self.n_spec)})  # rate constants
        pars.update({'intercept': 'Intercept'})  # intercept
        return pars

    def get_param_dicts(self):
        rates_dict = {f'k_{i + 1}': i for i in range(self.n_spec)}
        amps_dict = {name: i for i, name in enumerate(self.spec_names[:self.n_spec])}
        j_dict = {f'_{name}_0': i for i, name in enumerate(self.spec_names[:self.n_spec])}

        return dict(rates_dict=rates_dict, amps_dict=amps_dict, j_dict=j_dict)

    def add_params(self, param_set: list = None, dict_params: dict = None, j_dict: dict = None, rates_dict: dict = None,
                   amps_dict: dict = None, species_visible: [dict] = None, par_format=lambda name: name, **kwargs):
        has_been = False
        for par in param_set:
            f_par_name = par_format(par)

            vary = True
            value = 1
            min = -np.inf
            if self.is_j_par(par):
                dict_params['j'][j_dict[par]] = f_par_name
                vary = False
                if self.sequential:
                    value = 0 if has_been else 1
                    has_been = True

            elif self.is_rate_par(par):
                dict_params['rates'][rates_dict[par]] = f_par_name
                min = 0

            elif self.is_amp_par(par):
                dict_params['amps'][amps_dict[par]] = f_par_name
                vary = not self.varpro

                if species_visible is not None and par in species_visible:
                    value = 1 if species_visible[par] else 0
                else:
                    value = 1
            elif self.is_intercept(par):
                dict_params['intercept'] = f_par_name
                vary = self.fit_intercept_varpro and not self.varpro
                value = 0
            else:
                value = 0

            dict_params['all'].append(f_par_name)
            if f_par_name not in self.params:
                self.params.add(f_par_name, min=min, max=np.inf, value=value, vary=vary)
                # add an enabled attribute for each parameter
                self.params[f_par_name].enabled = True

    # @staticmethod
    # @ray.remote
    # def _get_results_remote(*views, rates=None, amps=None, js=None, intercepts=None, batch=None, visibles=None, **kwds):
    #     # view to data arrays has to be passed as individual arguments, list of argument will fail to use in numpy ops
    #
    #     results = []
    #
    #     for view, ks, j, vis, amps, icept, i in zip(views, rates, js, visibles, amps, intercepts, batch):
    #         amps = np.asarray(amps.copy(), dtype=np.float64)  # copy amps and change the dtype to float64
    #         n = ks.shape[0]
    #
    #         t = view[:, 0]
    #         y = view[:, 1]
    #
    #         if kwds['sequential']:
    #             # build the transfer matrix for sequential model
    #             K = np.zeros((n, n))
    #             for idx in range(n):
    #                 K[idx, idx] = -ks[idx]
    #                 if idx < n - 1:
    #                     K[idx + 1, idx] = ks[idx]
    #             # calculate and return the target model simulation
    #             traces = target_1st_order(t, K, j)
    #
    #         else:  # parallel model
    #             traces = parallel_model(j[None, :], t[:, None], ks[None, :])
    #
    #         _y = y.copy()
    #
    #         if kwds['varpro']:
    #
    #             # exp indep traces
    #             exp_dep_select = []
    #             exp_indep_select = []
    #             for key, visible in vis.items():
    #                 is_independent = key in kwds['exp_indep_amps']
    #                 exp_indep_select.append(is_independent and visible)
    #                 exp_dep_select.append(not is_independent and visible)
    #
    #             A = traces[:, exp_dep_select]  # select only visible species
    #             # add intercept as constant function
    #
    #             fit = 0
    #             if kwds['lstsq_intercept']:
    #                 A = np.hstack((A, np.ones_like(t)[:, None]))
    #             else:
    #                 fit = icept
    #                 _y -= fit
    #
    #             if any(exp_indep_select):  # calculate traces for independent-exp amplitudes and add to fit
    #                 exp_indep_traces = traces[:, exp_indep_select].dot(amps[exp_indep_select])  # add calculated traces
    #                 fit += exp_indep_traces
    #                 _y -= exp_indep_traces
    #
    #             # solve the least squares problem, find the amplitudes of visible compartments based on data
    #             _amps = OLS_ridge(A, _y, 0)  # A @ amps = y - A_fixed @ amps_fixes - intercept
    #
    #             fit += A.dot(_amps)  # calculate the fit and add it
    #
    #             # update amplitudes and intercept
    #             if kwds['lstsq_intercept']:
    #                 *__amps, icept = list(_amps)
    #                 amps[exp_dep_select] = __amps
    #
    #         else:
    #             fit = traces.dot(amps)  # weight the simulated traces with amplitudes and calculate the fit
    #
    #             if kwds['lstsq_intercept']:
    #                 icept = (y - fit).sum() / fit.shape[0]  # calculate intercept by least squares
    #
    #             fit += icept  # just add it to fit
    #
    #         results.append([i, fit, amps, icept])
    #
    #     return results

    def _get_traces(self, t, ks, j, i):

        n = self.n_spec

        assert n == ks.shape[0]

        if self.sequential:
            # build the transfer matrix for sequential model
            K = np.zeros((n, n))
            for idx in range(n):
                K[idx, idx] = -ks[idx]
                if idx < n - 1:
                    K[idx + 1, idx] = ks[idx]
            # calculate and return the target model simulation
            return target_1st_order(t, K, j)

        else:  # parallel model
            return parallel_model(j[None, :], t[:, None], ks[None, :])


class _FixedParametersModel(_InterceptVarProModel):

    def update_model_options(self, **kwargs):
        # handle the individual model option changes / override the default implementation
        for key, value in kwargs.items():
            if not hasattr(self, key):
                raise TypeError(f'Argument {key} is not valid.')
            self.__setattr__(key, value)

        if self.exp_dep_params is None:
            self.exp_dep_params = self.default_exp_dep_params()

        if len(kwargs) > 0:
            self.update_params()

        # also handles visible changes
        for exp_params, visible in zip(self.param_names_dict, self.spec_visible):
            for amp, vis in zip((self.params[name] for name in exp_params['amps']), visible.values()):
                amp.vary = not self.varpro and vis
                amp.enabled = not self.varpro and vis
                if not vis:
                    amp.value = 0

            self.params[exp_params['intercept']].vary = not self.fit_intercept_varpro
            self.params[exp_params['intercept']].enabled = not self.fit_intercept_varpro

        if 'intercept' in self.exp_indep_params:
            self.params[self.param_names_dict[0]['intercept']].enabled = True

        for par_name in self.exp_indep_params:
            if self.is_amp_par(par_name):
                p = self.params[par_name]
                p.vary = True
                p.enabled = True

    def get_current_species_names(self):
        return self.spec_names

    def _get_traces(self, t, ks, j, i):
        raise NotImplementedError()


class VarOrderModel(_FixedParametersModel):

    name = 'Variable Order A→B model'

    def __init__(self, *args, **kwargs):
        exps_data = args[0]  # exps_data must be first
        spec_names = ['A', 'B']
        spec_visible = [{'A': True, 'B': False} for _ in range(len(exps_data))]
        kwargs.update(n_spec=2, spec_names=spec_names, spec_visible=spec_visible)
        super(VarOrderModel, self).__init__(*args, **kwargs)

    def get_all_param_names(self):
        pars = {f'_{name}_0': f'Initial concentration of {name}' for name in self.spec_names}
        pars.update({name: f'Amplitude of {name}' for name in self.spec_names})  # amplitudes
        pars.update({'k': "The rate constant"})
        pars.update({'n': "Order of the reaction"})
        pars.update({'intercept': 'Intercept'})  # intercept
        return pars

    def default_exp_dep_params(self):
        pars = self.spec_names[:]  # amplitudes  # make a copy!!!
        pars += ['intercept']  # intercepts
        return pars

    def get_param_dicts(self):
        rates_dict = {'k': 0}
        amps_dict = {name: i for i, name in enumerate(self.spec_names)}
        j_dict = {f'_{name}_0': i for i, name in enumerate(self.spec_names)}

        return dict(rates_dict=rates_dict, amps_dict=amps_dict, j_dict=j_dict)

    def add_params(self, param_set: list = None, dict_params: dict = None, j_dict: dict = None, rates_dict: dict = None,
                   amps_dict: dict = None, species_visible: [dict] = None, par_format=lambda name: name, **kwargs):

        for par in param_set:
            f_par_name = par_format(par)

            vary = True
            value = 1
            min = -np.inf
            max = np.inf
            if self.is_j_par(par):
                dict_params['j'][j_dict[par]] = f_par_name
                vary = False
                if par == '_A_0':
                    value = 1
                else:
                    value = 0
            elif par.startswith('k'):
                dict_params['rates'][rates_dict[par]] = f_par_name
                min = 0
            elif self.is_amp_par(par):
                dict_params['amps'][amps_dict[par]] = f_par_name
                vary = not self.varpro
            elif self.is_intercept(par):
                dict_params['intercept'] = f_par_name
                vary = self.fit_intercept_varpro and not self.varpro
                value = 0
            elif par.startswith('n'):
                value = 1
                vary = True
                min = 0
                max = 20
            else:
                value = 0

            dict_params['all'].append(f_par_name)
            if f_par_name not in self.params:
                self.params.add(f_par_name, min=min, max=max, value=value, vary=vary)
                # add an enabled attribute for each parameter
                self.params[f_par_name].enabled = True

    def _get_traces(self, t, ks, j, i):

        j_A, j_B = j
        k = ks[0]

        n_par = list(filter(lambda name: name.startswith('n'), self.param_names_dict[i]['all']))[0]
        n = self.params[n_par].value

        C = np.empty((t.shape[0], 2), dtype=np.float64)

        if n == 1:  # first order
            C[:, 0] = j_A * np.exp(-k * t)
        else:  # n-th order, this wont work for negative x values, c(t) = 1-n root of (c^(1-n) + k*(n-1)*t)
            expr_in_root = np.power(float(j_A), 1 - n) + k * (n - 1) * t
            expr_in_root = expr_in_root.clip(min=0)  # set to 0 all the negative values
            C[:, 0] = np.power(expr_in_root, 1.0 / (1 - n))

        C[:, 1] = j_B + j_A - C[:, 0]

        return np.nan_to_num(np.heaviside(t[:, None], 1) * C)

class SingletOxygenRate(_FixedParametersModel):

    name = 'Singlet Oxygen Rate Constant'

    def __init__(self, *args, **kwargs):
        exps_data = args[0]  # exps_data must be first
        spec_names = ['A', 'B']
        spec_visible = [{'A': True, 'B': False} for _ in range(len(exps_data))]
        kwargs.update(n_spec=2, spec_names=spec_names, spec_visible=spec_visible)
        super(SingletOxygenRate, self).__init__(*args, **kwargs)

    def get_all_param_names(self):
        pars = {f'_{name}_0': f'Initial concentration of {name}' for name in self.spec_names}
        pars.update({name: f'Amplitude of {name}' for name in self.spec_names})  # amplitudes
        pars.update({'k': "The rate constant of singlet oxygen bimolecular reaction = kr"})
        pars.update({'rkd': "Singlet oxygen decay rate constant"})
        pars.update({'intercept': 'Intercept'})  # intercept
        return pars

    def default_exp_dep_params(self):
        pars = self.spec_names[:]  # amplitudes  # make a copy!!!
        pars += ['intercept']  # intercepts
        return pars

    def get_param_dicts(self):
        rates_dict = {'k': 0}
        amps_dict = {name: i for i, name in enumerate(self.spec_names)}
        j_dict = {f'_{name}_0': i for i, name in enumerate(self.spec_names)}

        return dict(rates_dict=rates_dict, amps_dict=amps_dict, j_dict=j_dict)

    def add_params(self, param_set: list = None, dict_params: dict = None, j_dict: dict = None, rates_dict: dict = None,
                   amps_dict: dict = None, species_visible: [dict] = None, par_format=lambda name: name, **kwargs):

        for par in param_set:
            f_par_name = par_format(par)

            vary = True
            value = 1
            min = -np.inf
            max = np.inf
            if self.is_j_par(par):
                dict_params['j'][j_dict[par]] = f_par_name
                vary = False
                if par == '_A_0':
                    value = 1
                else:
                    value = 0
            elif par.startswith('k'):
                dict_params['rates'][rates_dict[par]] = f_par_name
                min = 0
            elif self.is_amp_par(par):
                dict_params['amps'][amps_dict[par]] = f_par_name
                vary = not self.varpro
            elif self.is_intercept(par):
                dict_params['intercept'] = f_par_name
                vary = self.fit_intercept_varpro and not self.varpro
                value = 0
            elif par.startswith('rkd'):
                value = 1e6
                vary = False
                min = 0
                max = np.inf
            else:
                value = 0

            dict_params['all'].append(f_par_name)
            if f_par_name not in self.params:
                self.params.add(f_par_name, min=min, max=max, value=value, vary=vary)
                # add an enabled attribute for each parameter
                self.params[f_par_name].enabled = True

    def _get_traces(self, t, ks, j, i):

        j_A, j_B = j
        k = ks[0]

        n_par = list(filter(lambda name: name.startswith('rkd'), self.param_names_dict[i]['all']))[0]
        kd = self.params[n_par].value

        C = np.empty((t.shape[0], 2), dtype=np.float64)

        C[:, 0] = odeint(lambda c, _: -k * c / (kd + k * c), j_A, t).squeeze()
        C[:, 1] = j_B + j_A - C[:, 0]

        return np.nan_to_num(np.heaviside(t[:, None], 1) * C)


class Photosensitization(_FixedParametersModel):

    name = 'Photosensitization (PS→, PS+Q→T, T→) [Q]_0 \u226b [PS]_0'

    def __init__(self, *args, **kwargs):
        exps_data = args[0]  # exps_data must be first
        spec_names = ['PS', 'T']
        spec_visible = [{'PS': True, 'T': True} for _ in range(len(exps_data))]
        kwargs.update(n_spec=2, spec_names=spec_names, spec_visible=spec_visible)
        super(Photosensitization, self).__init__(*args, **kwargs)

    def get_all_param_names(self):
        pars = {'_Q_0': 'Concentration of a quencher'}
        pars.update({f'_{name}_0': f'Initial concentration of {name}' for name in self.spec_names})
        pars.update({name: f'Amplitude of {name}' for name in self.spec_names})  # amplitudes
        pars.update({'k_q': "Quenching rate constant"})
        pars.update({'k_PS': "Decay rate constant of a PS without a quencher"})
        pars.update({'k_T': "Decay rate constant of a sensitized triplet T"})
        pars.update({'intercept': 'Intercept'})  # intercept
        return pars

    def default_exp_dep_params(self):
        pars = ['intercept']  # intercepts
        pars += ['_Q_0']  # initial concentration of a quencher
        pars += self.spec_names  # amplitudes
        return pars

    def get_param_dicts(self):
        rates_dict = {name: i for i, name in enumerate(['k_q', 'k_PS', 'k_T'])}
        amps_dict = {name: i for i, name in enumerate(self.spec_names)}
        j_dict = {'_Q_0': 0}
        j_dict.update({f'_{name}_0': i+1 for i, name in enumerate(self.spec_names)})

        return dict(rates_dict=rates_dict, amps_dict=amps_dict, j_dict=j_dict)

    def add_params(self, param_set: list = None, dict_params: dict = None, j_dict: dict = None, rates_dict: dict = None,
                   amps_dict: dict = None, species_visible: [dict] = None, par_format=lambda name: name, **kwargs):

        for par in param_set:
            f_par_name = par_format(par)

            vary = True
            value = 1
            min = -np.inf
            if self.is_j_par(par):
                dict_params['j'][j_dict[par]] = f_par_name
                vary = False
                if par == '_PS_0' or par == '_Q_0':
                    value = 1
                else:
                    value = 0
            elif self.is_rate_par(par):
                dict_params['rates'][rates_dict[par]] = f_par_name
                min = 0
            elif self.is_amp_par(par):
                dict_params['amps'][amps_dict[par]] = f_par_name
                vary = not self.varpro
            elif self.is_intercept(par):
                dict_params['intercept'] = f_par_name
                vary = self.fit_intercept_varpro and not self.varpro
                value = 0
            else:
                value = 0

            dict_params['all'].append(f_par_name)
            if f_par_name not in self.params:
                self.params.add(f_par_name, min=min, max=np.inf, value=value, vary=vary)
                # add an enabled attribute for each parameter
                self.params[f_par_name].enabled = True

    def _get_traces(self, t, ks, j, i):

        assert ks.shape == j.shape

        # diff equations:
        #   dPS/dt = -k_PS * [PS] - k_q * [Q]*[PS]
        #   dT/dt = + k_q * [Q]*[PS] - k_T * [T]
        #
        #   solution for [Q]_0 >> [PS]_0
        #  c(PS) = [PS]_0 * exp(-(k_PS + k_q * [Q]) * t)
        #  c(T) = ... double exponential

        C = np.empty((t.shape[0], 2), dtype=np.float64)

        j_Q, j_PS, j_T = j
        k_q, k_PS, k_T = ks

        k_form = k_q * j_Q  # formation of T rate constant

        C[:, 0] = j_PS * np.exp(-t * (k_PS + k_form))
        C[:, 1] = j_T + (k_form / (k_T - k_form)) * (np.exp(-k_form * t) - np.exp(-k_T * t))

        return np.nan_to_num(np.heaviside(t[:, None], 1) * C)


class GeneralFitModel(_InterceptVarProModel):

    name = 'General model'

    def __init__(self, *args, show_backward_rates=False, **kwargs):
        super(GeneralFitModel, self).__init__(*args, **kwargs)

        self.general_model = GeneralModel()

        self.show_backward_rates = show_backward_rates

    def update_model_options(self, **kwargs):
        super(GeneralFitModel, self).update_model_options(**kwargs)

        # handle the individual model option changes / override the default implementation
        # also handles visible changes
        for exp_params, visible in zip(self.param_names_dict, self.spec_visible):
            for amp, vis in zip((self.params[name] for name in exp_params['amps']), visible.values()):
                cond = not self.varpro and vis and amp.name not in self.exp_indep_params
                amp.vary = cond
                amp.enabled = cond
                if not vis:
                    amp.value = 0

            self.params[exp_params['intercept']].vary = not self.fit_intercept_varpro
            self.params[exp_params['intercept']].enabled = not self.fit_intercept_varpro

        if 'intercept' in self.exp_indep_params:
            self.params[self.param_names_dict[0]['intercept']].enabled = True

        for par_name in self.exp_indep_params:
            if self.is_amp_par(par_name):
                p = self.params[par_name]
                p.vary = True
                p.enabled = True

    def get_all_param_names(self):
        pars = {}

        comps = self.general_model.get_compartments()

        # initial conditions
        pars.update({f'_{name}_0': f'Initial concentration of {name}' for name in comps})
        pars.update({name: f' Amplitude of {name}' for name in comps})  # amplitudes
        pars.update({rate: f'Rate constant {rate}' for rate in
                     self.general_model.get_rates(self.show_backward_rates, False)})  # rate constants
        pars.update({'intercept': 'Intercept'})  # intercept
        return pars

    def get_current_species_names(self):
        return self.general_model.get_compartments()

    def default_exp_dep_params(self):
        """Returns the set of default experiment-dependent parameters."""
        pars = self.general_model.get_compartments()
        pars += ['intercept']  # intercept
        return pars

    def load_from_file(self, fpath: str):
        self.general_model = GeneralModel.load(fpath)
        self.build()

    def load_from_scheme(self, scheme: str):
        self.general_model = GeneralModel.from_text(scheme)
        self.build()

    def build(self):
        self.general_model.build_func()
        comps = self.general_model.get_compartments()
        self.update_model_options(n_spec=len(comps), spec_names=comps)

    def __getattr__(self, item):
        return self.general_model.__getattribute__(item)

    def get_param_dicts(self):
        rate_vals = self.general_model.get_rates(self.show_backward_rates, True)
        rates_dict = {rate[0]: i for i, rate in enumerate(rate_vals)}
        amps_dict = {name: i for i, name in enumerate(self.spec_names)}
        j_dict = {f'_{name}_0': i for i, name in enumerate(self.spec_names)}

        return dict(rates_dict=rates_dict, amps_dict=amps_dict, j_dict=j_dict, rate_vals=rate_vals)

    def add_params(self, param_set: list = None, dict_params: dict = None, j_dict: dict = None, rates_dict: dict = None,
                   amps_dict: dict = None, species_visible: [dict] = None, par_format=lambda name: name, rate_vals=None):

        for par in param_set:
            f_par_name = par_format(par)

            vary = True
            value = 1
            min = -np.inf
            if self.is_j_par(par):
                dict_params['j'][j_dict[par]] = f_par_name  # place at correct position
                try:
                    # get species name from j format = remove initial _ sign and last 2 characters
                    value = self.general_model.initial_conditions[par[1:-2]]
                except KeyError:
                    pass
                vary = False
            elif self.is_rate_par(par):
                dict_params['rates'][rates_dict[par]] = f_par_name

                f = list(filter(lambda r: par == r[0], rate_vals))  # get rates from general model
                if len(f) > 0:
                    value = f[0][1]
                min = 0
            elif self.is_amp_par(par):
                dict_params['amps'][amps_dict[par]] = f_par_name
                vary = not self.varpro
            elif self.is_intercept(par):
                dict_params['intercept'] = f_par_name
                vary = self.fit_intercept_varpro and not self.varpro
                value = 0
            else:
                value = 0

            dict_params['all'].append(f_par_name)
            if f_par_name not in self.params:
                self.params.add(f_par_name, min=min, max=np.inf, value=value, vary=vary)
                # add an enabled attribute for each parameter
                self.params[f_par_name].enabled = True

    def _get_traces(self, x_data, rates, j, i):
        if x_data[0] > 0:  # initial conditions are valid for time=0
            n = 100  # prepend x values with 100 points if not starting with zero time
            x_prepended = np.concatenate((np.linspace(0, x_data[0], n, endpoint=False), x_data))
            return odeint(self.general_model.func, j, x_prepended, args=(rates,))[n:, :]

        elif x_data[0] < 0:
            x_pos = x_data[x_data >= 0]  # find x >= 0
            sol = np.zeros((x_data.shape[0], j.shape[0]), dtype=np.float64)
            if x_pos.shape[0] > 1:  # simulate only for at least 2 positive values
                sol[(x_data < 0).sum():, :] = self._get_traces(x_pos, rates, j, i)  # use recursion here

            return sol

        # for x_data[0] == 0
        return odeint(self.general_model.func, j, x_data, args=(rates,))

    def get_rate_values(self, exp_num=0):
        ks = np.asarray([self.params[p].value for p in self.param_names_dict[exp_num]['rates']])

        # transform ks to 2d array, needed for general model, first column are forward rates, the second backward rates
        if self.show_backward_rates:
            rates = np.empty((ks.shape[0] // 2, 2))
            rates[:, 0] = ks[::2]
            rates[:, 1] = ks[1::2]
        else:
            rates = np.vstack((ks, np.zeros(ks.shape[0]))).T

        return rates


# class LinearModel(_Model):
#     """This is the linear model,
#     :math:`y(x) = ax + y_0`
#     where :math:`a` is a slope and :math:`y_0` intercept.
#     """
#     name = 'Linear'
#
#     def init_params(self):
#         self.params.add('a', value=1, min=-np.inf, max=np.inf, vary=True)
#         self.params.add('y0', value=1, min=-np.inf, max=np.inf, vary=True)
#
#     def initialize_values(self, x_data, y_data):
#         self.params['a'].value = (y_data[-1] - y_data[0]) / (x_data[-1] - x_data[0])
#         self.params['y0'].value = y_data[0]
#
#     @staticmethod
#     def func(x, a, y0):
#         return a * x + y0
#
#
# class Photobleaching_Model(_Model):
#     """This is the photobleaching model. Differential equation:
#
#     :math:`\\frac{\mathrm{d}c(t)}{\mathrm{d}t} = -k\\left(1 - 10^{-\\varepsilon c(t)l}\\right)`
#
#     where :math:`k = \\frac{1}{V}I_0\Phi` is a product of incident light intensity and photoreaction quantum yield
#     divided by the volume of the solution.
#
#     initial conditions: :math:`c(0) = c_0 = \\frac{A_0}{\\varepsilon l}`
#
#     Integrated law for evolution of absorbance in time (fitting equation):
#
#     :math:`A(t) = \\frac{1}{\\ln 10}\\ln\\left(  10^{-A_0k^{\\prime}t}  \\left( 10^{A_0} + 10^{A_0k^{\\prime}t} - 1  \\right)\\right)`
#
#     where :math:`A_0` is absorbance at time 0, :math:`k^{\\prime} = \\frac{k}{c_0}`, both :math:`k` and :math:`c_0`
#     are fitting parameters, one of them must be fixed, usually :math:`c_0`
#     """
#     name = 'Photobleaching'
#
#     def init_params(self):
#         self.params.add('c0', value=1, min=0, max=np.inf, vary=False)
#         self.params.add('A0', value=1, min=-np.inf, max=np.inf, vary=True)
#         self.params.add('k', value=1, min=0, max=np.inf, vary=True)
#         self.params.add('y0', value=0, min=-np.inf, max=np.inf, vary=True)
#
#     def initialize_values(self, x_data, y_data):
#         self.params['A0'].value = y_data[0]
#
#     @staticmethod
#     def func(x, c0, A0, k, y0):
#         kp = k / c0  # k'
#         return np.log(10 ** (-A0 * kp * x) * (10 ** A0 + 10 ** (A0 * kp * x) - 1)) / np.log(10) + y0
#
#
# class Eyring(_Model):
#     """This is the Eyring model,
#     :math:`k(T) = \\frac{\\kappa k_B T}{h}e^{\\frac{\\Delta S}{R}}e^{-\\frac{\\Delta H}{RT}}`
#     """
#     name = 'Eyring'
#
#     def init_params(self):
#         self.params.add('dH', value=1000, min=-np.inf, max=np.inf, vary=True)
#         self.params.add('dS', value=100, min=-np.inf, max=np.inf, vary=True)
#         self.params.add('kappa', value=1, min=0, max=1, vary=False)
#
#     def initialize_values(self, x_data, y_data):
#         # self.params['a'].value = (y_data[-1] - y_data[0]) / (x_data[-1] - x_data[0])
#         # self.params['y0'].value = y_data[0]
#         pass
#
#     @staticmethod
#     def func(x, dH, dS, kappa):
#         kB = 1.38064852 * 1e-23  # boltzmann
#         h = 6.62607004 * 1e-34  # planck
#         R = 8.31446261815324  # universal gas constant
#         # x - thermodynamic temperature
#         return (kappa * kB * x / h) * np.exp((dS * x - dH) / (R * x))



import os
import sys
from abc import abstractmethod


# abstract class that every parser should inherit
class Parser:

    def __init__(self, filepath=None, str_data=None, delimiter='\t', decimal_sep='.'):
        if filepath is None and str_data is None:
            raise ValueError("At least one argument (filepath or str_data) must not be None.")

        if str_data is None:
            self.filepath = filepath
            self.name_of_file = self.get_filename()
            self._data = None
        else:
            self._data = str_data
            self.filepath = None
            self.name_of_file = ''

        self.delimiter = delimiter
        self.decimal_sep = decimal_sep

    @abstractmethod
    def parse(self, name=None):
        # it = self.__iter__()  # get iterator

        # for line in it:
        #     # Do stuff...
        raise NotImplementedError("This method has to be implemented.")

    def __iter__(self):
        """Return an line iterator that iterates through loaded data, either opened file, or string data"""
        if self._data is None:

            encoding = self.get_encoding()

            if encoding is None:
                raise Exception("Problem with encoding of the file.\n{}".format(self.filepath))

            return open(self.filepath, 'r', encoding=encoding, newline='').__iter__()
        else:
            return iter(self._data.splitlines(keepends=True))

    def get_encoding(self):
        if self.filepath is None:
            return None

        default_encoding = sys.getfilesystemencoding()
        # utf-8-sig is better, since it does not matter if BOM is on the begining of the file or not
        encoding = 'utf-8-sig' if default_encoding == 'utf-8' else default_encoding
        try:
            with open(self.filepath, 'r', encoding=encoding) as f:
                f.readline()
            return encoding
        except UnicodeDecodeError:
            encoding = 'utf-16'
            try:
                with open(self.filepath, 'r', encoding=encoding) as f:
                    f.readline()
                return encoding
            except Exception as ex:
                # Logger.console_message(ex.__str__())
                return None
        except Exception as ex:
            # Logger.console_message(ex.__str__())
            return None

    def get_filename(self):
        # head = os.path.split(path_to_file)[0]
        tail = os.path.split(self.filepath)[1]

        return os.path.splitext(tail)[0]  # without extension

    def float_try_parse(self, num):
        try:
            return float(num.replace(self.decimal_sep, '.').strip())
        except ValueError:
            return None





# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'spectramanipulator/dialogs\stylewidget_gui.ui'
#
# Created by: PyQt5 UI code generator 5.15.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(419, 334)
        self.gridLayout_2 = QtWidgets.QGridLayout(Form)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.combSymbol = QtWidgets.QComboBox(Form)
        self.combSymbol.setObjectName("combSymbol")
        self.gridLayout.addWidget(self.combSymbol, 6, 2, 1, 2)
        self.label_7 = QtWidgets.QLabel(Form)
        self.label_7.setObjectName("label_7")
        self.gridLayout.addWidget(self.label_7, 8, 0, 1, 1)
        self.label_6 = QtWidgets.QLabel(Form)
        self.label_6.setObjectName("label_6")
        self.gridLayout.addWidget(self.label_6, 7, 0, 1, 1)
        self.btnSymBrushColor = QtWidgets.QPushButton(Form)
        self.btnSymBrushColor.setAutoFillBackground(False)
        self.btnSymBrushColor.setObjectName("btnSymBrushColor")
        self.gridLayout.addWidget(self.btnSymBrushColor, 7, 2, 1, 1)
        self.btnSymFillColor = QtWidgets.QPushButton(Form)
        self.btnSymFillColor.setAutoFillBackground(False)
        self.btnSymFillColor.setObjectName("btnSymFillColor")
        self.gridLayout.addWidget(self.btnSymFillColor, 8, 2, 1, 1)
        self.sbSymBrushAlpha = QtWidgets.QSpinBox(Form)
        self.sbSymBrushAlpha.setMaximum(255)
        self.sbSymBrushAlpha.setObjectName("sbSymBrushAlpha")
        self.gridLayout.addWidget(self.sbSymBrushAlpha, 7, 3, 1, 1)
        self.cbSymBrushDefault = QtWidgets.QCheckBox(Form)
        self.cbSymBrushDefault.setObjectName("cbSymBrushDefault")
        self.gridLayout.addWidget(self.cbSymBrushDefault, 7, 4, 1, 1)
        self.cbSymFillDefault = QtWidgets.QCheckBox(Form)
        self.cbSymFillDefault.setObjectName("cbSymFillDefault")
        self.gridLayout.addWidget(self.cbSymFillDefault, 8, 4, 1, 1)
        self.sbSymFillAlpha = QtWidgets.QSpinBox(Form)
        self.sbSymFillAlpha.setMaximum(255)
        self.sbSymFillAlpha.setObjectName("sbSymFillAlpha")
        self.gridLayout.addWidget(self.sbSymFillAlpha, 8, 3, 1, 1)
        self.label_5 = QtWidgets.QLabel(Form)
        self.label_5.setObjectName("label_5")
        self.gridLayout.addWidget(self.label_5, 6, 0, 1, 1)
        self.label = QtWidgets.QLabel(Form)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 3, 0, 1, 1)
        self.btnColor = QtWidgets.QPushButton(Form)
        self.btnColor.setAutoFillBackground(False)
        self.btnColor.setObjectName("btnColor")
        self.gridLayout.addWidget(self.btnColor, 3, 2, 1, 1)
        self.sbAlpha = QtWidgets.QSpinBox(Form)
        self.sbAlpha.setMaximum(255)
        self.sbAlpha.setObjectName("sbAlpha")
        self.gridLayout.addWidget(self.sbAlpha, 3, 3, 1, 1)
        self.cbColor = QtWidgets.QCheckBox(Form)
        self.cbColor.setObjectName("cbColor")
        self.gridLayout.addWidget(self.cbColor, 3, 4, 1, 1)
        self.cbPlotLegend = QtWidgets.QCheckBox(Form)
        self.cbPlotLegend.setObjectName("cbPlotLegend")
        self.gridLayout.addWidget(self.cbPlotLegend, 10, 0, 1, 1)
        self.cbLineWidth = QtWidgets.QCheckBox(Form)
        self.cbLineWidth.setObjectName("cbLineWidth")
        self.gridLayout.addWidget(self.cbLineWidth, 4, 4, 1, 1)
        self.dsbLineWidth = QtWidgets.QDoubleSpinBox(Form)
        self.dsbLineWidth.setDecimals(1)
        self.dsbLineWidth.setSingleStep(0.1)
        self.dsbLineWidth.setObjectName("dsbLineWidth")
        self.gridLayout.addWidget(self.dsbLineWidth, 4, 2, 1, 2)
        self.cbLineType = QtWidgets.QCheckBox(Form)
        self.cbLineType.setObjectName("cbLineType")
        self.gridLayout.addWidget(self.cbLineType, 5, 4, 1, 1)
        self.label_3 = QtWidgets.QLabel(Form)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 5, 0, 1, 1)
        self.label_4 = QtWidgets.QLabel(Form)
        self.label_4.setObjectName("label_4")
        self.gridLayout.addWidget(self.label_4, 4, 0, 1, 1)
        self.label_2 = QtWidgets.QLabel(Form)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 2, 3, 1, 1)
        self.combLineType = QtWidgets.QComboBox(Form)
        self.combLineType.setObjectName("combLineType")
        self.gridLayout.addWidget(self.combLineType, 5, 2, 1, 2)
        self.label_8 = QtWidgets.QLabel(Form)
        self.label_8.setObjectName("label_8")
        self.gridLayout.addWidget(self.label_8, 9, 0, 1, 1)
        self.dsbSymSize = QtWidgets.QDoubleSpinBox(Form)
        self.dsbSymSize.setDecimals(1)
        self.dsbSymSize.setSingleStep(0.1)
        self.dsbSymSize.setObjectName("dsbSymSize")
        self.gridLayout.addWidget(self.dsbSymSize, 9, 2, 1, 2)
        self.verticalLayout.addLayout(self.gridLayout)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.btnSetDefauls = QtWidgets.QPushButton(Form)
        self.btnSetDefauls.setAutoFillBackground(False)
        self.btnSetDefauls.setObjectName("btnSetDefauls")
        self.horizontalLayout.addWidget(self.btnSetDefauls)
        self.btnCancel = QtWidgets.QPushButton(Form)
        self.btnCancel.setAutoFillBackground(False)
        self.btnCancel.setObjectName("btnCancel")
        self.horizontalLayout.addWidget(self.btnCancel)
        self.btnApply = QtWidgets.QPushButton(Form)
        self.btnApply.setAutoFillBackground(False)
        self.btnApply.setObjectName("btnApply")
        self.horizontalLayout.addWidget(self.btnApply)
        self.btnOK = QtWidgets.QPushButton(Form)
        self.btnOK.setAutoFillBackground(False)
        self.btnOK.setObjectName("btnOK")
        self.horizontalLayout.addWidget(self.btnOK)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.gridLayout_2.addLayout(self.verticalLayout, 0, 0, 1, 1)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label_7.setText(_translate("Form", "Symbol fill color"))
        self.label_6.setText(_translate("Form", "Symbol brush color"))
        self.btnSymBrushColor.setText(_translate("Form", "Pick color"))
        self.btnSymFillColor.setText(_translate("Form", "Pick color"))
        self.cbSymBrushDefault.setText(_translate("Form", "Set same as line"))
        self.cbSymFillDefault.setText(_translate("Form", "Set same as line"))
        self.label_5.setText(_translate("Form", "Symbol type"))
        self.label.setText(_translate("Form", "Color"))
        self.btnColor.setText(_translate("Form", "Pick color"))
        self.cbColor.setText(_translate("Form", "Default"))
        self.cbPlotLegend.setText(_translate("Form", "Plot legend"))
        self.cbLineWidth.setText(_translate("Form", "Default"))
        self.cbLineType.setText(_translate("Form", "Default"))
        self.label_3.setText(_translate("Form", "Line type"))
        self.label_4.setText(_translate("Form", "Line width"))
        self.label_2.setText(_translate("Form", "Alpha"))
        self.label_8.setText(_translate("Form", "Symbol size"))
        self.btnSetDefauls.setText(_translate("Form", "Set defaults"))
        self.btnCancel.setText(_translate("Form", "Cancel"))
        self.btnApply.setText(_translate("Form", "Apply"))
        self.btnOK.setText(_translate("Form", "OK"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    Form = QtWidgets.QWidget()
    ui = Ui_Form()
    ui.setupUi(Form)
    Form.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'spectramanipulator/dialogs\gui_interpolate_dialog.ui'
#
# Created by: PyQt5 UI code generator 5.15.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(356, 203)
        self.layoutWidget = QtWidgets.QWidget(Dialog)
        self.layoutWidget.setGeometry(QtCore.QRect(20, 20, 321, 171))
        self.layoutWidget.setObjectName("layoutWidget")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.layoutWidget)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(self.layoutWidget)
        self.label.setWordWrap(True)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.label_2 = QtWidgets.QLabel(self.layoutWidget)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 0, 0, 1, 1)
        self.label_3 = QtWidgets.QLabel(self.layoutWidget)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 1, 0, 1, 1)
        self.leSpacing = QtWidgets.QLineEdit(self.layoutWidget)
        self.leSpacing.setObjectName("leSpacing")
        self.gridLayout.addWidget(self.leSpacing, 0, 2, 1, 1)
        self.cbIntepolation = QtWidgets.QComboBox(self.layoutWidget)
        self.cbIntepolation.setObjectName("cbIntepolation")
        self.gridLayout.addWidget(self.cbIntepolation, 1, 2, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        self.verticalLayout_2.addLayout(self.verticalLayout)
        self.buttonBox = QtWidgets.QDialogButtonBox(self.layoutWidget)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout_2.addWidget(self.buttonBox)

        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Dialog"))
        self.label.setText(_translate("Dialog", "<html><head/><body><p>Set the spacing of x values and kind of interpolation. For documentation, see <a href=\"https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.interp1d.html#scipy.interpolate.interp1d\"><span style=\" text-decoration: underline; color:#0000ff;\">scipy.interpolate.interp1d.</span></a> Eg. for resampling the unevenly spaced spectral data by 1 nm, set spacing to 1.</p></body></html>"))
        self.label_2.setText(_translate("Dialog", "Spacing"))
        self.label_3.setText(_translate("Dialog", "Interpolation"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    Dialog = QtWidgets.QDialog()
    ui = Ui_Dialog()
    ui.setupUi(Dialog)
    Dialog.show()
    sys.exit(app.exec_())



from PyQt5 import QtCore

from PyQt5.QtWidgets import QDockWidget, QWidget

from qtconsole.rich_jupyter_widget import RichJupyterWidget
from qtconsole.inprocess import QtInProcessKernelManager
import spectramanipulator


# great help from here https://stackoverflow.com/questions/11513132/embedding-ipython-qt-console-in-a-pyqt-application/12375397#12375397
# just copied, works like charm :)

# qtconsole package is not yet compatible with pyqt6

class ConsoleWidget(RichJupyterWidget):

    def __init__(self, customBanner=None, *args, **kwargs):
        super(ConsoleWidget, self).__init__(*args, **kwargs)

        if customBanner is not None:
            self.banner = customBanner

        self.font_size = 6
        self.kernel_manager = kernel_manager = QtInProcessKernelManager()
        kernel_manager.start_kernel(show_banner=False)
        kernel_manager.kernel.gui = 'qt'
        self.kernel_client = kernel_client = self._kernel_manager.client()
        kernel_client.start_channels()

        def stop():
            kernel_client.stop_channels()
            kernel_manager.shutdown_kernel()
            guisupport.get_app_qt().exit()

        self.exit_requested.connect(stop)

    def set_focus(self):
        # TODO--- does not work...
        self._control.setFocus()

    def push_vars(self, variableDict):
        """
        Given a dictionary containing name / value pairs, push those variables
        to the Jupyter console widget
        """
        self.kernel_manager.kernel.shell.push(variableDict)

    def clear(self):
        """
        Clears the terminal
        """
        self._control.clear()

        # self.kernel_manager

    def print_text(self, text):
        """
        Prints some plain text to the console
        """
        # self._append_plain_text(text)
        self.append_stream(text)

    def print_html(self, text):
        self._append_html(text, True)

    def execute_command(self, command, hidden=True):
        """
        Execute a command in the frame of the console widget
        """
        self._execute(command, hidden)


class Console(QDockWidget):
    """
    Console window used for advanced commands, debugging,
    logging, and profiling.
    """

    _instance = None

    def __init__(self, parentWindow):
        QDockWidget.__init__(self, parentWindow)
        self.setTitleBarWidget(QWidget())
        self.setAllowedAreas(QtCore.Qt.BottomDockWidgetArea)
        self.setFeatures(QDockWidget.NoDockWidgetFeatures)
        self.setVisible(False)

        banner = f"""Simple Spectra Manipulator console based on IPython, version {spectramanipulator.__version__}. Numpy package was imported as np and matplotlib.pyplot as plt. Three variables are setup:
    item - this is used to interact with spectra in TreeWidget and perform various calculations
    tree_widget - instance of TreeWidget
    main - instance of Main (Main Window)

Enjoy.
        
"""
        Console._instance = self

        # Add console window
        self.console_widget = ConsoleWidget(banner)
        self.setWidget(self.console_widget)

    def setVisible(self, visible):
        """
        Override to set proper focus.
        """
        QDockWidget.setVisible(self, visible)
        if visible:
            self.console_widget.setFocus()

    @staticmethod
    def showMessage(message, add_new_line=True):
        if Console._instance is not None:
            string = '\n' + message if add_new_line else message
            Console._instance.console_widget.print_text(string)

    def print_html(self, text, add_new_line=True):
        string = '\n' + text if add_new_line else text
        self.console_widget.print_html(string)

    @staticmethod
    def execute_command(cmd, hidden=True):
        if Console._instance is not None:
            Console._instance.console_widget.execute_command(cmd, hidden)

    @staticmethod
    def push_variables(variable_dict):
        if Console._instance is not None:
            Console._instance.console_widget.push_vars(variable_dict)




from pyqtgraph.graphicsItems.LegendItem import LegendItem as _LegendItem
from pyqtgraph.graphicsItems.LegendItem import ItemSample as _ItemSamle
from .label_item import LabelItem

from PyQt5 import QtGui, QtCore
from pyqtgraph import functions as fn
from pyqtgraph.graphicsItems.ScatterPlotItem import ScatterPlotItem, drawSymbol
from pyqtgraph.graphicsItems.PlotDataItem import PlotDataItem
from pyqtgraph.graphicsItems.BarGraphItem import BarGraphItem

# changed legend line position in ItemSample


class LegendItem(_LegendItem):

    def __init__(self, size=None, offset=None, horSpacing=25, verSpacing=0,
                 pen=None, brush=None, labelTextColor=None, frame=True,
                 labelTextSize='9pt', colCount=1, sampleType=None, **kwargs):

        super(LegendItem, self).__init__(size, offset, horSpacing, verSpacing,
                                         pen, brush, labelTextColor, frame,
                                         labelTextSize, colCount, sampleType, **kwargs)

        self.verSpacing = verSpacing  # verSpacing parameter used as a setup for setMinimumHeight in LabelItemModif

    def remove_last_item(self):
        sample, label = self.items[-1]
        del self.items[-1]
        self.layout.removeItem(sample)  # remove from layout
        sample.close()  # remove from drawing
        self.layout.removeItem(label)
        label.close()
        self.updateSize()  # redraq box

    def addItem(self, item, name):
        """
        Add a new entry to the legend.

        ==============  ========================================================
        **Arguments:**
        item            A :class:`~pyqtgraph.PlotDataItem` from which the line
                        and point style of the item will be determined or an
                        instance of ItemSample (or a subclass), allowing the
                        item display to be customized.
        title           The title to display for this item. Simple HTML allowed.
        ==============  ========================================================
        """
        # USED LabelItemModif insted of LabelItem
        label = LabelItem(name, color=self.opts['labelTextColor'],
                          justify='left', size=self.opts['labelTextSize'], verspacing=self.verSpacing)
        if isinstance(item, ItemSample):  # Changed from ItemSample to ItemSampleModif
            sample = item
        else:
            sample = ItemSample(item)  # Changed from ItemSample to ItemSampleModif
        self.items.append((sample, label))
        self._addItemToLayout(sample, label)
        self.updateSize()


class ItemSample(_ItemSamle):

    def paint(self, p, *args):
        opts = self.item.opts

        if opts.get('antialias'):
            p.setRenderHint(p.Antialiasing)

        if not isinstance(self.item, ScatterPlotItem):
            p.setPen(fn.mkPen(opts['pen']))
            # p.drawLine(0, 11, 20, 11)
            p.drawLine(0, 15, 20, 15)  # CHANGED THIS LINE

            if (opts.get('fillLevel', None) is not None and
                    opts.get('fillBrush', None) is not None):
                p.setBrush(fn.mkBrush(opts['fillBrush']))
                p.setPen(fn.mkPen(opts['fillBrush']))
                p.drawPolygon(QtGui.QPolygonF(
                    [QtCore.QPointF(2, 18), QtCore.QPointF(18, 2),
                     QtCore.QPointF(18, 18)]))

        symbol = opts.get('symbol', None)
        if symbol is not None:
            if isinstance(self.item, PlotDataItem):
                opts = self.item.scatter.opts
            p.translate(10, 10)
            drawSymbol(p, symbol, opts['size'], fn.mkPen(opts['pen']),
                       fn.mkBrush(opts['brush']))

        if isinstance(self.item, BarGraphItem):
            p.setBrush(fn.mkBrush(opts['brush']))
            p.drawRect(QtCore.QRectF(2, 2, 18, 18))




import os
import numpy as np
from concurrent.futures import ProcessPoolExecutor
from functools import partial
# import time

from .parsers import GeneralParser, DXFileParser, CSVFileParser
from .settings import Settings
from .logger import Logger


def parse_files(filepaths, settings: dict = None):
    settings_dict = Settings.get_attributes()
    if settings is not None:
        settings_dict.update(settings)

    parse_func = partial(_process_filepath, settings=settings_dict)
    return _parse_files(filepaths, parse_func)


def parse_files_specific(filepaths, use_CSV_parser=False, **kwargs):
    """kwargs are passed to the parser"""

    parse_func = partial(_process_filepath_specific, use_CSV_parser=use_CSV_parser, **kwargs)
    return _parse_files(filepaths, parse_func)


def parse_text(text):
    txt_parser = GeneralParser(str_data=text, delimiter=Settings.clip_imp_delimiter,
                               decimal_sep=Settings.clip_imp_decimal_sep,
                               remove_empty_entries=Settings.remove_empty_entries,
                               skip_col_num=Settings.skip_columns_num,
                               general_import_spectra_name_from_filename=Settings.general_import_spectra_name_from_filename,
                               general_if_header_is_empty_use_filename=Settings.general_if_header_is_empty_use_filename,
                               skip_nan_columns=Settings.skip_nan_columns,
                               nan_replacement=Settings.nan_replacement)

    spectra = txt_parser.parse()
    if spectra is None or len(spectra) == 0:
        Logger.message("No lines were parsed, check delimiter, decimal separator and number of skipped columns in settings.")
    return spectra


def _process_filepath(filepath, settings):
    try:
        parsed_spectra, parser = _parse_file(filepath, settings)
        return None, parsed_spectra, parser
    except Exception as ex:
        return "Error occurred while parsing file {}, " \
               "skipping the file.\nException: {}\n".format(filepath, ex.__str__()), None, None


def _process_filepath_specific(filepath, use_CSV_parser=True, **kwargs):
    try:
        parser = CSVFileParser(filepath, **kwargs) if use_CSV_parser else GeneralParser(filepath, **kwargs)
        return None, parser.parse(), parser
    except Exception as ex:
        return "Error occurred while parsing file {}, " \
               "skipping the file.\nException: {}\n".format(filepath, ex.__str__()), None, None


def _parse_files(filepaths, parse_func):
    """
    parse_func takes filepath as argument and returns the tuple with error and result.

    :param filepaths:
    :param settings:
    :return:
    """

    if filepaths is None:
        raise ValueError("Filepaths cannot be None.")
    if isinstance(filepaths, str):
        filepaths = [filepaths]
    if not isinstance(filepaths, list):
        raise ValueError("Filepaths must be a list of strings")

    filesizes = np.asarray([os.path.getsize(fname) for fname in filepaths])
    spectra = []
    parsers = []

    # start_time = time.time()

    # if at least two files are >= 0.5 MB, switch to multiprocess
    if (os.cpu_count() > 1 and Settings.enable_multiprocessing and filesizes.shape[0] > 1 and
       (filesizes >= 0.5e6).sum() > 1) or Settings.force_multiprocess:

        # Logger.console_message("Loading in multiprocess.")

        with ProcessPoolExecutor() as executor:
            results = executor.map(parse_func, filepaths)

            for err, p_spectra, parser in results:
                if err is not None:
                    Logger.message(err)
                    continue

                if p_spectra is not None:
                    spectra += p_spectra
                    parsers.append(parser)
    else:
        # Logger.console_message("Loading in the main thread.")
        for filepath in filepaths:
            err, p_spectra, parser = parse_func(filepath)

            if err is not None:
                Logger.message(err)
                continue

            if p_spectra is not None:
                spectra += p_spectra
                parsers.append(parser)

    # elapsed_time = time.time() - start_time
    # Logger.console_message(elapsed_time)

    if len(spectra) != 0:
        return spectra, parsers
    else:
        Logger.message("No lines were parsed, check delimiter, decimal separator, number of skipped columns and"
                       " skip in settings.")
    return None, None


def _parse_file(filepath, settings):
    file, ext = os.path.splitext(filepath)
    ext = ext.lower()

    if ext == '.csv':
        txt_parser = CSVFileParser(filepath, delimiter=settings['csv_imp_delimiter'],
                                   decimal_sep=settings['csv_imp_decimal_sep'],
                                   remove_empty_entries=settings['remove_empty_entries'],
                                   skip_col_num=settings['skip_columns_num'],
                                   general_import_spectra_name_from_filename=settings['general_import_spectra_name_from_filename'],
                                   general_if_header_is_empty_use_filename=settings['general_if_header_is_empty_use_filename'],
                                   skip_nan_columns=settings['skip_nan_columns'],
                                   nan_replacement=settings['nan_replacement'])
        return txt_parser.parse(), txt_parser
    # DX file
    elif ext == '.dx':
        dx_parser = DXFileParser(filepath, delimiter=settings['dx_imp_delimiter'],
                                 decimal_sep=settings['dx_imp_decimal_sep'],
                                 dx_import_spectra_name_from_filename=settings['dx_import_spectra_name_from_filename'],
                                 dx_if_title_is_empty_use_filename=settings['dx_if_title_is_empty_use_filename'])
        return dx_parser.parse(), dx_parser
    # general (could be any file)
    else:
        txt_parser = GeneralParser(filepath, delimiter=settings['general_imp_delimiter'],
                                   decimal_sep=settings['general_imp_decimal_sep'],
                                   remove_empty_entries=settings['remove_empty_entries'],
                                   skip_col_num=settings['skip_columns_num'],
                                   general_import_spectra_name_from_filename=settings['general_import_spectra_name_from_filename'],
                                   general_if_header_is_empty_use_filename=settings['general_if_header_is_empty_use_filename'],
                                   skip_nan_columns=settings['skip_nan_columns'],
                                   nan_replacement=settings['nan_replacement'])

        return txt_parser.parse(), txt_parser



import numpy as np
from lmfit import fit_report, report_fit, report_ci, conf_interval, conf_interval2d

import matplotlib.pyplot as plt
from matplotlib.ticker import ScalarFormatter

from ..user_namespace import plot_fit


class FitResult(object):
    """Helper class that stores the results after fitting and provide additional functions such as printing
    or plotting confidence intervals.

    Attributes
    ----------

    result : :class:`MinimizerResult`
        Result from fitting as obtained by LMFIT.
    minimizer : :class:`Minimizer`
        Minimizer class used for fitting.
    params : :class:`Parameters`
        Variable that stored parameters as :class:`MinimizerResult` class
    values_errors : numpy.ndarray
        2D numpy array of fitted paramater values (first column) and corresponding errors (second column)
    model : :class:`_Model`
        Model used for fitting.
    """

    def __init__(self, minimizer_result, minimizer, values_errors, current_model,
                 data_item=None, fit_item=None, residuals_item=None):
        self.result = minimizer_result
        self.minimizer = minimizer
        self.params = minimizer_result.params
        self.values_errors = values_errors
        self.model = current_model

        self.data_item = data_item
        self.fit_item = fit_item
        self.residuals_item = residuals_item

    def report(self, print=True):
        """Prints a fit report if print == True, otherwise returns fit report as text.

        Parameters
        ----------
        print : bool
            If True, fit report is printed.
        """
        return report_fit(self.result) if print else fit_report(self.result)

    # https://lmfit.github.io/lmfit-py/confidence.html
    def confidence_intervals(self, p_names=None, sigmas=(1, 2, 3)):
        """Prints a confidence intervals.

        Parameters
        ----------
        p_names : {list, None}, optional
            Names of the parameters for which the confidence intervals are calculated. If None (default),
            the confidence intervals are calculated for every parameter.
        sigmas : {list, tuple}, optional
            The sigma-levels to find (default is [1, 2, 3]). See Note below.

        Note
        ----
        The values for sigma are taken as the number of standard deviations for a normal distribution
        and converted to probabilities. That is, the default sigma=[1, 2, 3] will use probabilities of
        0.6827, 0.9545, and 0.9973. If any of the sigma values is less than 1, that will be interpreted
        as a probability. That is, a value of 1 and 0.6827 will give the same results, within precision.

        """
        ci = conf_interval(self.minimizer, self.result, p_names=p_names, sigmas=sigmas)
        report_ci(ci)

    # https://lmfit.github.io/lmfit-py/confidence.html
    def confidence_interval2D(self, x_name, y_name, nx=10, ny=10):
        """Draws a 2D confidence intervals using matplotlib.

        Parameters
        ----------
        x_name : str
            Name of the variable that will be on the x axis.
        y_name : str
            Name of the variable that will be on the y axis.
        nx : int, optional
            Number of points in the x direction, default 10, the higher the value, better resolution, but slower.
        ny : int, optional
            Number of points in the y direction, default 10, the higher the value, better resolution, but slower.
        """
        cx, cy, grid = conf_interval2d(self.minimizer, self.result, x_name, y_name, nx, ny)
        plt.contourf(cx, cy, grid, np.linspace(0, 1, 21))
        plt.xlabel(x_name)
        plt.colorbar()
        plt.ylabel(y_name)
        plt.show()

    def plot(self, symlog=False, linscale=1, linthresh=100,
                  lw_data=0.5, lw_fit=1.5, fig_size=(5, 4), y_label='$\\Delta$A', x_label='Time / $\\mu$s',
                  x_lim=(None, None), x_major_formatter=ScalarFormatter(), x_major_locator=None, y_major_locator=None,
             data_color='red', t_mul_factor=1, y_lim=(None, None), filepath=None, dpi=500, transparent=False):

        plot_fit(self.data_item, self.fit_item, self.residuals_item, symlog=symlog, linscale=linscale,
                  linthresh=linthresh, lw_data=lw_data, lw_fit=lw_fit, fig_size=fig_size, y_label=y_label,
                  x_label=x_label, x_lim=x_lim, t_mul_factor=t_mul_factor, y_lim=y_lim, filepath=filepath,
                  dpi=dpi, transparent=transparent, x_major_formatter=x_major_formatter,
                  x_major_locator=x_major_locator, y_major_locator=y_major_locator, data_color=data_color)





from PyQt5 import QtGui
import pyqtgraph as pg

from spectramanipulator.pyqtgraphmodif.legend_item import LegendItem
from spectramanipulator.pyqtgraphmodif.plot_item import PlotItem
from spectramanipulator.pyqtgraphmodif.view_box import ViewBox

from pyqtgraph.exporters import ImageExporter
from pyqtgraph.exporters import SVGExporter

from spectramanipulator.settings import Settings

from PyQt5.QtCore import Qt
# from PyQt5.QtGui import QColor


# subclassing of PlotWidget
class PlotWidget(pg.PlotWidget):
    _instance = None

    def __init__(self, parent=None, coordinates_func=None):

        pg.setConfigOption('background', 'w')
        pg.setConfigOption('foreground', 'k')

        self.plotted_items = {}  # dictionary with keys as spectrum objects and values as plot data items
        self.plotted_fits = {}  #

        self.plotItem = PlotItem(viewBox=ViewBox(self.plotted_items))

        # use our modified plotItem
        super(PlotWidget, self).__init__(parent, plotItem=self.plotItem)

        PlotWidget._instance = self

        self.coordinates_func = coordinates_func

        self.legend = None
        # self.add_legend(spacing=Settings.legend_spacing)

        self.img_exporter = ImageExporter(self.plotItem)
        self.svg_exporter = SVGExporter(self.plotItem)

        self.lr_item = None  # linear region item

        self.update_settings()
        self.plotItem.setDownsampling(ds=True, auto=True, mode='subsample')
        self.plotItem.setClipToView(True)

    def update_settings(self):

        self.clear_plots()
        self.add_legend(spacing=Settings.legend_spacing)

        pg.setConfigOptions(antialias=Settings.antialiasing)
        self.plotItem.showAxis('top', show=True)
        self.plotItem.showAxis('right', show=True)

        self.plotItem.setTitle(Settings.graph_title, size="{}pt".format(Settings.graph_title_font_size))

        self.plotItem.setLabel('left', text=Settings.left_axis_label, units=Settings.left_axis_unit)
        self.plotItem.setLabel('bottom', text=Settings.bottom_axis_label, units=Settings.bottom_axis_unit)
        self.plotItem.showGrid(x=Settings.show_grid, y=Settings.show_grid, alpha=Settings.grid_alpha)

        left_label_font = QtGui.QFont()
        left_label_font.setPixelSize(Settings.left_axis_font_size)

        self.plotItem.getAxis('left').label.setFont(left_label_font)

        bottom_label_font = QtGui.QFont()
        bottom_label_font.setPixelSize(Settings.bottom_axis_font_size)

        self.plotItem.getAxis('bottom').label.setFont(bottom_label_font)

    def plot(self, item, **kwargs):
        """kwargs are passed to plotItem.plot function"""

        # if spectrum is already plotted, update the data and style
        if item in self.plotted_items:
            self.plotted_items[item].setData(item.data, **kwargs)
            if 'zValue' in kwargs:
                self.plotted_items[item].setZValue(kwargs['zValue'])
            return

        pi = self.plotItem.plot(item.data, **kwargs)
        self.plotted_items[item] = pi

    def update_items_data(self, items: list):
        """Updates the plots of items."""
        for item in items:
            try:
                self.plotted_items[item].setData(item.data)
            except KeyError:
                continue

    def remove(self, spectrum):
        if spectrum in self.plotted_items:
            self.removeItem(self.plotted_items[spectrum])  # remove from pyqtgraph
            del self.plotted_items[spectrum]  # remove entry in dictionary

    def clear_plots(self):
        """Removes all plots except of linear region item and fits"""
        # self.plotted_spectra.clear()
        for item in self.plotted_items.values():
            self.removeItem(item)

        self.plotted_items.clear()

    @classmethod
    def plotted_items(cls):
        if cls._instance is not None:
            return cls._instance.plotted_items

    @classmethod
    def plot_fit(cls, item, **kwargs):
        self = cls._instance
        if self is None:
            return

        if item in self.plotted_fits:
            self.plotted_fits[item].setData(item.data, **kwargs)
            if 'zValue' in kwargs:
                self.plotted_fits[item].setZValue(kwargs['zValue'])
            return

        pi = self.plotItem.plot(item.data, **kwargs)
        self.plotted_fits[item] = pi

    @classmethod
    def remove_fits(cls, items: list):
        self = cls._instance
        if self is None:
            return

        for item in items:
            if item in self.plotted_fits:
                self.removeItem(self.plotted_fits[item])
                del self.plotted_fits[item]

    @classmethod
    def remove_all_fits(cls):
        self = cls._instance
        if self is None:
            return

        for val in self.plotted_fits.values():
            self.removeItem(val)

        self.plotted_fits.clear()

    @classmethod
    def add_linear_region(cls, region=None, bounds=None, brush=None, orientation='vertical', z_value=-10):
        """bounds is tuple or None"""

        self = cls._instance
        if self is None:
            return

        if region is None:
            # if no region is setup, sets a region from 13 - 87 % of the current view range
            region = self.getViewBox().viewRange()[0]
            f = 0.13
            diff = region[1] - region[0]
            region[0] += f * diff
            region[1] -= f * diff

        brush = QtGui.QBrush(QtGui.QColor(0, 255, 0, 20)) if brush is None else brush

        if self.lr_item is None:
            self.lr_item = pg.LinearRegionItem(region, orientation=orientation, brush=brush, bounds=bounds)
            self.addItem(self.lr_item)

        self.lr_item.setRegion(region)
        if bounds is not None:
            self.lr_item.setBounds(bounds)
        self.lr_item.setBrush(brush)
        self.lr_item.setZValue(z_value)

        return self.lr_item

    @classmethod
    def remove_linear_region(cls):
        self = cls._instance
        if self is None:
            return

        if self.lr_item is not None:
            self.removeItem(self.lr_item)
            self.lr_item = None

    @classmethod
    def set_lr_x_range(cls, x0, x1):
        self = cls._instance
        if self is None:
            return

        if self.lr_item is not None:
            self.lr_item.setRegion((x0, x1))

    @classmethod
    def get_view_range(cls):
        if cls._instance is None:
            return

        x0, x1 = cls._instance.getViewBox().viewRange()[0]
        y0, y1 = cls._instance.viewRange()[1]

        return x0, x1, y0, y1

    def save_plot_to_clipboard_as_png(self):
        self.img_exporter.export(copy=True)

    def save_plot_to_clipboard_as_svg(self):
        self.svg_exporter.export(copy=True)

    def leaveEvent(self, ev):
        super(PlotWidget, self).leaveEvent(ev)
        self.setCursor(Qt.ArrowCursor)

    def mouseMoveEvent(self, ev):
        super(PlotWidget, self).mouseMoveEvent(ev)

        pos = ev.pos()

        in_scene = self.plotItem.sceneBoundingRect().contains(pos)
        in_legend = self.legend.sceneBoundingRect().contains(pos)

        in_lin_reg = False
        in_lin_reg_line = False
        lin_reg_moving = False
        if self.lr_item is not None:
            in_lin_reg = self.lr_item.sceneBoundingRect().contains(pos)
            line1, line2 = self.lr_item.lines
            in_lin_reg_line = line1.sceneBoundingRect().contains(pos) or line2.sceneBoundingRect().contains(pos)
            lin_reg_moving = self.lr_item.moving or line1.moving or line2.moving

        if in_scene:
            try:
                mouse_point = self.plotItem.vb.mapSceneToView(pos)
                n = Settings.coordinates_sig_figures
                # double format with n being the number of significant figures of a number
                self.coordinates_func(f"x={{:.{n}g}}, y={{:.{n}g}}".format(mouse_point.x(), mouse_point.y()))
            except:
                pass

        # set the corresponding cursor
        if in_scene and not lin_reg_moving:
            if in_lin_reg_line:
                self.viewport().setCursor(Qt.SizeHorCursor)
            elif in_legend or in_lin_reg:
                self.viewport().setCursor(Qt.SizeAllCursor)
            else:
                self.viewport().setCursor(Qt.CrossCursor)

    def add_legend(self, size=None, spacing=5, offset=(-30, 30)):
        if self.legend is not None:
            self.legend.clear()
            self.plotItem.legend = None
            del self.legend

        self.legend = LegendItem(size, verSpacing=spacing, offset=offset)
        self.legend.setParentItem(self.plotItem.vb)
        self.plotItem.legend = self.legend






import sys
import os
from spectramanipulator.settings import Settings
import winshell
from ctypes import windll

# --- Window control ---

SW_SHOW = 5    # activate and display
SW_SHOWNA = 8  # show without activation
SW_HIDE = 0

GetConsoleWindow = windll.kernel32.GetConsoleWindow
ShowWindow = windll.user32.ShowWindow
IsWindowVisible = windll.user32.IsWindowVisible

# Handle to console window associated with current Python
# interpreter procss, 0 if there is no window
console_window_handle = GetConsoleWindow()


def set_attached_console_visible(state):
    """Show/hide system console window attached to current process.
       Return it's previous state.

       Availability: Windows"""
    flag = {True: SW_SHOW, False: SW_HIDE}
    return bool(ShowWindow(console_window_handle, flag[state]))


def is_attached_console_visible():
    """Return True if attached console window is visible"""
    return IsWindowVisible(console_window_handle)


def set_windows_appusermodelid():
    """Make sure correct icon is used on Windows 7 taskbar"""
    try:
        return windll.shell32.SetCurrentProcessExplicitAppUserModelID("spyder.Spyder")
    except AttributeError:
        return "SetCurrentProcessExplicitAppUserModelID not found"


# ---- Registry ----

ROOT_PATH = r"Software\Classes"
REG_PATH_EXT = f"{ROOT_PATH}\\{Settings.PROJECT_EXTENSION}"  # \.smpj
HKEY_CURRENT_USER_text = "HKEY_CURRENT_USER"


def associate_project_file():
    if sys.platform != 'win32':
        print(f"Association of {Settings.PROJECT_EXTENSION} project files files only implemented for Windows systems.")
        return

    import winreg as reg

    # help from https://stackoverflow.com/questions/1387769/create-registry-entry-to-associate-file-extension-with-application-in-c

    # [HKEY_CURRENT_USER\Software\Classes\blergcorp.blergapp.v1\shell\open\command]
    # @= "c:\path\to\app.exe \"%1\""
    # [HKEY_CURRENT_USER\Software\Classes\.blerg]
    # @= "blergcorp.blergapp.v1"

    try:
        REG_PATH_PROGRAM = f"{ROOT_PATH}\\{Settings.REG_PROGRAM_NAME}\\shell\\open\\command"

        reg.CreateKey(reg.HKEY_CURRENT_USER, REG_PATH_EXT)
        registry_key_ext = reg.OpenKey(reg.HKEY_CURRENT_USER, REG_PATH_EXT, 0, reg.KEY_WRITE)
        reg.SetValue(registry_key_ext, '', reg.REG_SZ, Settings.REG_PROGRAM_NAME)
        reg.CloseKey(registry_key_ext)

        print(f"Added {HKEY_CURRENT_USER_text}\\{REG_PATH_EXT}")

        # curr_dirr = os.path.dirname(os.path.realpath(__file__))
        # prog_path = os.path.join(curr_dirr, "spectramanipulator.pyw")
        prog_path = "spectramanipulator"

        py_exec = sys.executable  # find python executable
        # if py_exec.endswith('python.exe'):
        #     py_exec = os.path.join(os.path.dirname(sys.executable), 'pythonw.exe')  # use pythonw.exe

        # executable = f"\"{py_exec}\" \"{prog_path}\" \"%1\""  # path is "[pythonw.exe]" -m "[ssm.pyw]" "[path-to-file-arg]"
        executable = f"\"{py_exec}\" -m {prog_path} \"%1\""  # path is "[python.exe]" -m spectramanipulator "[path-to-file-arg]"
        # print(executable)

        reg.CreateKey(reg.HKEY_CURRENT_USER, REG_PATH_PROGRAM)
        registry_key_prog = reg.OpenKey(reg.HKEY_CURRENT_USER, REG_PATH_PROGRAM, 0, reg.KEY_WRITE)
        reg.SetValue(registry_key_prog, '', reg.REG_SZ, executable)
        reg.CloseKey(registry_key_prog)

        print(f"Added {HKEY_CURRENT_USER_text}\\{REG_PATH_PROGRAM}")

    except Exception as ex:
        print(ex.__str__())


def _delete_subkey(key0: int, key1: str, key2=""):
    # from https://stackoverflow.com/questions/38205784/python-how-to-delete-registry-key-and-subkeys-from-hklm-getting-error-5
    import winreg as reg

    if key2 == "":
        currentkey = key1
    else:
        currentkey = f"{key1}\\{key2}"

    open_key = reg.OpenKey(key0, currentkey, 0, reg.KEY_ALL_ACCESS)
    infokey = reg.QueryInfoKey(open_key)
    for x in range(0, infokey[0]):
        #  NOTE:: This code is to delete the key and all subkeys.
        #  If you just want to walk through them, then
        #  you should pass x to EnumKey. subkey = _winreg.EnumKey(open_key, x)
        #  Deleting the subkey will change the SubKey count used by EnumKey.
        #  We must always pass 0 to EnumKey so we
        #  always get back the new first SubKey.
        subkey = reg.EnumKey(open_key, 0)
        try:
            reg.DeleteKey(open_key, subkey)
            print(f"Removed {HKEY_CURRENT_USER_text}\\{currentkey}\\{subkey}")
        except:
            _delete_subkey(key0, currentkey, subkey)
            # no extra delete here since each call
            # to deleteSubkey will try to delete itself when its empty.

    reg.DeleteKey(open_key, "")
    open_key.Close()
    print(f"Removed {HKEY_CURRENT_USER_text}\\{currentkey}")


def remove_project_file_association():
    if sys.platform != 'win32':
        print(f"Association of {Settings.PROJECT_EXTENSION} project files files only implemented for Windows systems.")
        return

    import winreg as reg

    try:
        REG_PATH_PROGRAM = f"{ROOT_PATH}\\{Settings.REG_PROGRAM_NAME}"

        _delete_subkey(reg.HKEY_CURRENT_USER, REG_PATH_EXT)
        _delete_subkey(reg.HKEY_CURRENT_USER, REG_PATH_PROGRAM)

    except Exception as ex:
        print(ex.__str__())


def create_desktop_shortcut():
    if sys.platform != 'win32':
        print(f"Creation of shortcut only works for Windows systems.")
        return

    """Creates a desktop shortcut, help from https://pbpython.com/windows-shortcut.html"""
    path = os.path.join(winshell.desktop(), "SM.lnk")
    with winshell.shortcut(path) as shortcut:
        shortcut.path = sys.executable
        shortcut.description = "Spectra Manipulator"
        shortcut.arguments = "-m spectramanipulator"
        # shortcut.icon_location = (icon, 0)
        shortcut.working_directory = os.path.dirname(sys.executable)


if __name__ == '__main__':
    if len(sys.argv) > 1 and sys.argv[1] == '--remove':
        remove_project_file_association()
        print(f"\nProject file {Settings.PROJECT_EXTENSION} association was removed from win registry.")
    else:
        associate_project_file()
        print(f"\nProject file {Settings.PROJECT_EXTENSION} association was added to win registry.")







from PyQt5 import QtWidgets

from PyQt5.QtWidgets import QLineEdit, QComboBox
from spectramanipulator.dialogs.genericinputdialog import GenericInputDialog

from webbrowser import open_new_tab


class TrustRegionReflOptionDialog(GenericInputDialog):

    def __init__(self, ftol=1e-8, xtol=1e-8, gtol=1e-8, loss: str = 'linear',
                 max_nfev=None, verbose: int = 2, set_result=None):

        if self.is_opened:
            return

        # https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.least_squares.html

        self.options = self.default_opts()

        self.options.update(ftol=ftol, xtol=xtol, gtol=gtol, loss=loss,
                            max_nfev=max_nfev, verbose=verbose)

        widget_list = []

        self.ftol_edit = QLineEdit(str(ftol))
        self.xtol_edit = QLineEdit(str(xtol))
        self.gtol_edit = QLineEdit(str(gtol))

        # self.minim_alg_opts = [
        #     {'opt': 'trf', 'description': 'Trust Region Reflective algorithm, particularly suitable for large sparse problems with bounds. Generally robust method.'},
        #     {'opt': 'dogbox', 'description': 'Dogleg algorithm with rectangular trust regions, typical use case is small problems with bounds. Not recommended for problems with rank-deficient Jacobian.'},
        #     {'opt': 'lm', 'description': 'Levenberg-Marquardt algorithm as implemented in MINPACK. Doesn’t handle bounds and sparse Jacobians. Usually the most efficient method for small unconstrained problems.'},
        # ]

        self.loss_options = [
            {'opt': 'linear', 'description': 'rho(z) = z. Gives a standard least-squares problem.'},
            {'opt': 'soft_l1', 'description': 'rho(z) = 2 * ((1 + z)**0.5 - 1). The smooth approximation of l1 (absolute value) loss. Usually a good choice for robust least squares.'},
            {'opt': 'huber', 'description': 'rho(z) = z if z <= 1 else 2*z**0.5 - 1. Works similarly to ‘soft_l1’.'},
            {'opt': 'cauchy', 'description': 'rho(z) = ln(1 + z). Severely weakens outliers influence, but may cause difficulties in optimization process.'},
            {'opt': 'arctan', 'description': 'rho(z) = arctan(z). Limits a maximum loss on a single residual, has properties similar to ‘cauchy’.'},
        ]

        self.verbose_opts = [
            {'opt': 0, 'description': 'Works silently'},
            {'opt': 1, 'description': 'Displays a termination report.'},
            {'opt': 2, 'description': 'Displays progress during iterations (not supported by ‘lm’ method)'},
        ]

        # self.alg_cb = QComboBox()
        # self.alg_cb.addItems(map(lambda o: o['description'], self.minim_alg_opts))
        # _opts = list(map(lambda o: o['opt'], self.minim_alg_opts))
        # self.alg_cb.setCurrentIndex(_opts.index(method))

        self.loss_cb = QComboBox()
        self.loss_cb.addItems(map(lambda o: f"{o['opt']}: {o['description']}", self.loss_options))
        _opts = list(map(lambda o: o['opt'], self.loss_options))
        self.loss_cb.setCurrentIndex(_opts.index(loss))

        self.max_nfev_edit = QLineEdit(str(max_nfev))

        self.verbose_cb = QComboBox()
        self.verbose_cb.addItems(map(lambda o: f"{o['opt']}: {o['description']}", self.verbose_opts))
        _opts = list(map(lambda o: o['opt'], self.verbose_opts))
        self.verbose_cb.setCurrentIndex(_opts.index(verbose))

        # widget_list.append(('Algorithm to perform minimization:', self.alg_cb))
        widget_list.append(('ftol:', self.ftol_edit))
        widget_list.append(('xtol:', self.xtol_edit))
        widget_list.append(('gtol:', self.gtol_edit))
        widget_list.append(('Loss function:', self.loss_cb))
        widget_list.append(('Max n_fev:', self.max_nfev_edit))
        widget_list.append(('Verbose:', self.verbose_cb))

        description = """<html><head/><body><p>For explanation, see 
        <a href="https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.least_squares.html">
        <span style=" text-decoration: underline; color:#0000ff;">scipy.optimize.least_squares</span></a>.</p></body></html>"""

        super(TrustRegionReflOptionDialog, self).__init__(widget_list=widget_list,
                                                          label_text=description,
                                                          title='Settings of Trust Region Reflective algorithm',
                                                          parent=None,
                                                          set_result=set_result)

        self.label.linkActivated.connect(lambda: open_new_tab("https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.least_squares.html"))


    @staticmethod
    def default_opts():
        return dict(ftol=1e-8, xtol=1e-8, gtol=1e-8, loss='linear',
                    max_nfev=None, verbose=2)

    def accept(self):
        # self.options['method'] = self.minim_alg_opts[self.alg_cb.currentIndex()]['opt']
        self.options['loss'] = self.loss_options[self.loss_cb.currentIndex()]['opt']
        self.options['verbose'] = self.verbose_opts[self.verbose_cb.currentIndex()]['opt']
        try:
            self.options['ftol'] = float(self.ftol_edit.text())
        except ValueError:
            self.options['ftol'] = None
        try:
            self.options['xtol'] = float(self.xtol_edit.text())
        except ValueError:
            self.options['xtol'] = None
        try:
            self.options['gtol'] = float(self.gtol_edit.text())
        except ValueError:
            self.options['gtol'] = None
        try:
            self.options['max_nfev'] = float(self.max_nfev_edit.text())
        except ValueError:
            self.options['max_nfev'] = None

        super(TrustRegionReflOptionDialog, self).accept()


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    Dialog = TrustRegionReflOptionDialog()
    Dialog.show()
    sys.exit(app.exec_())




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'spectramanipulator/dialogs\gui_intintinputdialog.ui'
#
# Created by: PyQt5 UI code generator 5.15.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(459, 146)
        self.buttonBox = QtWidgets.QDialogButtonBox(Dialog)
        self.buttonBox.setGeometry(QtCore.QRect(370, 60, 71, 71))
        self.buttonBox.setOrientation(QtCore.Qt.Vertical)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayoutWidget = QtWidgets.QWidget(Dialog)
        self.verticalLayoutWidget.setGeometry(QtCore.QRect(20, 20, 331, 101))
        self.verticalLayoutWidget.setObjectName("verticalLayoutWidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.verticalLayoutWidget)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(self.verticalLayoutWidget)
        self.label.setWordWrap(True)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.label_3 = QtWidgets.QLabel(self.verticalLayoutWidget)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 1, 1, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem, 0, 0, 1, 1)
        self.label_2 = QtWidgets.QLabel(self.verticalLayoutWidget)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 0, 1, 1, 1)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem1, 1, 0, 1, 1)
        self.sbOffset = QtWidgets.QSpinBox(self.verticalLayoutWidget)
        self.sbOffset.setObjectName("sbOffset")
        self.gridLayout.addWidget(self.sbOffset, 1, 3, 1, 1)
        self.sbn = QtWidgets.QSpinBox(self.verticalLayoutWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.sbn.sizePolicy().hasHeightForWidth())
        self.sbn.setSizePolicy(sizePolicy)
        self.sbn.setObjectName("sbn")
        self.gridLayout.addWidget(self.sbn, 0, 3, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)

        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)
        QtCore.QMetaObject.connectSlotsByName(Dialog)
        Dialog.setTabOrder(self.sbn, self.sbOffset)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Dialog"))
        self.label.setText(_translate("Dialog", "<html><head/><body><p>label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label label </p></body></html>"))
        self.label_3.setText(_translate("Dialog", "shift"))
        self.label_2.setText(_translate("Dialog", "n"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    Dialog = QtWidgets.QDialog()
    ui = Ui_Dialog()
    ui.setupUi(Dialog)
    Dialog.show()
    sys.exit(app.exec_())





from PyQt5.QtCore import QMimeData
from pickle import dumps, loads


class PyMimeData(QMimeData):
    """ The PyMimeData wraps a Python instance as MIME data.
    """
    # The MIME type for instances.
    MIME_TYPE = 'application/x-ssm-dragged_items'

    def __init__(self, item_list=None):
        """ Initialise the instance.
        """
        QMimeData.__init__(self)
        # print('init PyMimeData')

        # Keep a local reference to be returned if possible.
        self._local_instance = item_list

        if item_list is not None:
            # We may not be able to pickle the data.
            try:
                pdata = dumps(item_list)
            except Exception as ex:
                print("dumps not successfull", ex.__str__())
                return

            # This format (as opposed to using a single sequence) allows the
            # type to be extracted without unpickling the data itself.
            # self.setData(self.MIME_TYPE, dumps(item_list.__class__) + pdata)

            self.setData(self.MIME_TYPE, pdata)

    @classmethod
    def coerce(cls, md):
        """ Coerce a QMimeData instance to a PyMimeData instance if
        possible.
        """
        # See if the data is already of the right type.  If it is then
        # print('coarse PyMimeData')

        if isinstance(md, cls):
            return md

        # See if the data type is supported.
        if not md.hasFormat(cls.MIME_TYPE):
            return None

        nmd = cls()
        nmd.setData(cls.MIME_TYPE, md.data(cls.MIME_TYPE))

        return nmd

    def instance(self):
        """ Return the instance.
        """
        if self._local_instance is not None:
            return self._local_instance

        # io = StringIO(str(self.data(self.MIME_TYPE)))

        try:
            items_list = loads(self.data(self.MIME_TYPE))

            self._local_instance = items_list

            # Recreate the instance.
            return items_list
        except:
            pass

        return None

        # io = StringIO(str(self.data(self.MIME_TYPE)))
        #
        # try:
        #     # Skip the type.
        #     load(io)
        #
        #     # Recreate the instance.
        #     return load(io)
        # except:
        #     pass
        #
        # return None

    # what is this for??
    def instanceType(self):
        """ Return the type of the instance.
        """
        if self._local_instance is not None:
            return self._local_instance.__class__

        try:
            return loads(str(self.data(self.MIME_TYPE)))
        except:
            pass

        return None